# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from datetime import datetime
from io import StringIO
from re import match
from typing import TYPE_CHECKING, Any, Generic, Literal, Optional, TypeVar, Union, overload

import sqlalchemy
from dogpile.cache.api import NoValue
from sqlalchemy.exc import DatabaseError, IntegrityError, OperationalError
from sqlalchemy.orm import aliased
from sqlalchemy.orm.exc import FlushError
from sqlalchemy.sql.expression import Executable, and_, delete, desc, false, func, or_, select, true

from rucio.common import exception, types, utils
from rucio.common.cache import MemcacheRegion
from rucio.common.checksum import CHECKSUM_KEY, GLOBALLY_SUPPORTED_CHECKSUMS
from rucio.common.config import get_lfn2pfn_algorithm_default
from rucio.common.constants import RSE_ALL_SUPPORTED_PROTOCOL_OPERATIONS, RSE_ATTRS_BOOL, RSE_ATTRS_STR, SUPPORTED_SIGN_URL_SERVICES_LITERAL, RseAttr
from rucio.common.utils import Availability
from rucio.core.rse_counter import add_counter, get_counter
from rucio.db.sqla import models
from rucio.db.sqla.constants import ReplicaState, RSEType
from rucio.db.sqla.session import read_session, stream_session, transactional_session
from rucio.db.sqla.util import temp_table_mngr

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from sqlalchemy.orm import Session
    from sqlalchemy.sql._typing import _ColumnsClauseArgument
    from typing_extensions import Self

T = TypeVar('T', bound="RseData")

RSE_SETTINGS = ["continent", "city", "region_code", "country_name", "time_zone", "ISP", "ASN"]
REGION = MemcacheRegion(expiration_time=900)


class RseData:
    """
    Helper data class storing rse data grouped in one place.
    """
    def __init__(
            self,
            id_: str,
            name: Optional[str] = None,
            columns: Optional[dict[str, Any]] = None,
            attributes: Optional[dict[str, Any]] = None,
            info: Optional[types.RSESettingsDict] = None,
            usage: Optional[list[dict[str, Any]]] = None,
            limits: Optional[dict[str, Any]] = None,
            transfer_limits: Optional[dict[str, Any]] = None
    ):
        self.id = id_
        self._name = name
        self._columns = columns
        self._attributes = attributes
        self._info = info
        self._usage = usage
        self._limits = limits
        self._transfer_limits = transfer_limits

    def _get_loaded_attribute(self, attribute_name: str) -> Any:
        attribute = getattr(self, f'_{attribute_name}', None)
        if attribute is None:
            raise ValueError(f'{attribute_name} not loaded for rse {self}')
        return attribute

    @property
    def name(self) -> str:
        return self._get_loaded_attribute('name')

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def columns(self) -> dict[str, Any]:
        return self._get_loaded_attribute('columns')

    @property
    def attributes(self) -> dict[str, Any]:
        return self._get_loaded_attribute('attributes')

    @property
    def info(self) -> types.RSESettingsDict:
        return self._get_loaded_attribute('info')

    @property
    def usage(self) -> list[dict[str, Any]]:
        return self._get_loaded_attribute('usage')

    @property
    def limits(self) -> dict[str, Any]:
        return self._get_loaded_attribute('limits')

    @property
    def transfer_limits(self) -> dict[str, Any]:
        return self._get_loaded_attribute('transfer_limits')

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        if self._name is not None:
            return self._name
        return self.id

    def __eq__(self, other) -> bool:
        if other is None:
            return False
        return self.id == other.id

    def is_tape(self) -> bool:
        if self.info['rse_type'] == RSEType.TAPE or self.info['rse_type'] == 'TAPE' or self.attributes.get(RseAttr.STAGING_REQUIRED, False):
            return True
        return False

    def is_tape_or_staging_required(self) -> bool:
        if self.is_tape() or self.attributes.get(RseAttr.STAGING_REQUIRED, False):
            return True
        return False

    @read_session
    def ensure_loaded(
        self,
        load_name: bool = False,
        load_columns: bool = False,
        load_attributes: bool = False,
        load_info: bool = False,
        load_usage: bool = False,
        load_limits: bool = False,
        load_transfer_limits: bool = False,
        *,
        session: "Session"
    ) -> "Self":
        if self._columns is None and load_columns:
            self._columns = get_rse(rse_id=self.id, session=session)
            self._name = self._columns['rse']
        if self._attributes is None and load_attributes:
            self._attributes = list_rse_attributes(self.id, use_cache=True, session=session)
        if self._info is None and load_info:
            self._info = get_rse_info(self.id, session=session)
            self._name = self._info['rse']
        if self._usage is None and load_usage:
            self._usage = get_rse_usage(rse_id=self.id, session=session)
        if self._limits is None and load_limits:
            self._limits = get_rse_limits(rse_id=self.id, session=session)
        if self._transfer_limits is None and load_transfer_limits:
            self._transfer_limits = get_rse_transfer_limits(rse_id=self.id, session=session)
        if self._name is None and load_name:
            self._name = get_rse_name(rse_id=self.id, session=session)
        return self

    @staticmethod
    @read_session
    def bulk_load(
            rse_id_to_data: "dict[str, RseData]",
            load_name: bool = False,
            load_columns: bool = False,
            load_attributes: bool = False,
            load_info: bool = False,
            load_usage: bool = False,
            load_limits: bool = False,
            include_deleted: bool = False,
            *,
            session: "Session"
    ) -> None:
        """
        Given a dict of RseData objects indexed by rse_id, ensure that the desired fields are initialised
        in all objects from the input.
        """
        if len(rse_id_to_data) < 4:  # 4 was selected without particular reason as "seems good enough"
            for rse_data in rse_id_to_data.values():
                rse_data.ensure_loaded(
                    load_name=load_name,
                    load_columns=load_columns,
                    load_attributes=load_attributes,
                    load_info=load_info,
                    load_usage=load_usage,
                    load_limits=load_limits,
                    session=session
                )
            return

        rse_ids_to_load = set()
        for rse_id, rse_data in rse_id_to_data.items():
            anything_to_load = any((
                load_name and rse_data._name is None,
                load_columns and rse_data._columns is None,
                load_attributes and rse_data._attributes is None,
                load_info and rse_data._info is None,
                load_usage and rse_data._usage is None,
                load_limits and rse_data._limits is None,
            ))
            if anything_to_load:
                rse_ids_to_load.add(rse_id)
        if not rse_ids_to_load:
            # all required fields are already present. Nothing to do
            return

        temp_table = temp_table_mngr(session).create_id_table()
        session.bulk_insert_mappings(temp_table, ({'id': rse_id} for rse_id in rse_ids_to_load))

        # We need to ensure that all rses exist and are not deleted. We could check this with a specialized
        # query, but this seems wasteful under normal operation: the caller of the current function probably
        # got the list of RSE IDs from list_rses (or another source which checks for deleted rses).
        #
        # Instead, directly fetch all RSEs, which allows to reduce the number (and complexity) of other queries below
        stmt = select(
            models.RSE
        ).join_from(
            temp_table,
            models.RSE,
            models.RSE.id == temp_table.id
        )
        if not include_deleted:
            stmt = stmt.where(
                models.RSE.deleted == false()
            )
        db_rses_by_id = {str(db_rse.id): db_rse for db_rse in session.execute(stmt).scalars()}

        if len(db_rses_by_id) != len(rse_ids_to_load):
            failed_rse_ids = ', '.join(rse_ids_to_load.difference(db_rses_by_id))
            raise exception.RSENotFound(f"RSE(s) with id(s) '{failed_rse_ids}' cannot be found")

        if load_attributes:
            for rse_id, attr in _fetch_many_rses_attributes(rse_id_temp_table=temp_table, session=session):
                rse_id_to_data[rse_id]._attributes = attr

        if load_columns:
            settings_by_id = {}
            if not load_attributes:
                settings_by_id = dict(_fetch_many_rses_attributes(rse_id_temp_table=temp_table,
                                                                  keys=RSE_SETTINGS,
                                                                  session=session))
            for rse_id, db_rse in db_rses_by_id.items():
                rse_data = rse_id_to_data[rse_id]
                settings = rse_data._attributes if rse_data._attributes is not None else settings_by_id.get(rse_id, {})
                columns = _format_get_rse(db_rse=db_rse, rse_attributes=settings, session=session)
                rse_data._columns = columns
                rse_data._name = columns['rse']

        if load_info:
            stmt = select(
                temp_table.id,
                models.RSEProtocol
            ).outerjoin_from(
                temp_table,
                models.RSEProtocol,
                models.RSEProtocol.rse_id == temp_table.id
            ).order_by(
                temp_table.id,
            )
            for rse_id, db_protocols in _group_query_result_by_rse_id(stmt, session=session):
                db_rse = db_rses_by_id[rse_id]
                rse_data = rse_id_to_data[rse_id]
                rse_attributes = rse_data._attributes
                info = _format_get_rse_protocols(rse=db_rse, db_protocols=db_protocols,
                                                 rse_attributes=rse_attributes, session=session)
                rse_data._info = info
                rse_data._name = info['rse']

        if load_limits:
            stmt = select(
                temp_table.id,
                models.RSELimit
            ).outerjoin_from(
                temp_table,
                models.RSELimit,
                models.RSELimit.rse_id == temp_table.id
            ).order_by(
                temp_table.id,
            )
            for rse_id, limits_list in _group_query_result_by_rse_id(stmt, session=session):
                rse_id_to_data[rse_id]._limits = {limit.name: limit.value for limit in limits_list}

        if load_usage:
            stmt = select(
                temp_table.id,
                models.RSEUsage
            ).outerjoin_from(
                temp_table,
                models.RSEUsage,
                models.RSEUsage.rse_id == temp_table.id
            ).order_by(
                temp_table.id,
            )
            for rse_id, db_usages in _group_query_result_by_rse_id(stmt, session=session):
                usage = _format_get_rse_usage(rse_id=rse_id, db_usages=db_usages, per_account=False, session=session)
                rse_id_to_data[rse_id]._usage = usage

        if load_name:
            # The name could have been loaded already (when loading columns or info). Skip loading if it's known.
            if not load_columns and not load_info:
                for rse_id in rse_ids_to_load:
                    rse_id_to_data[rse_id]._name = db_rses_by_id[rse_id].rse


class RseCollection(Generic[T]):
    """
    Container which keeps track of information loaded from the database for a group of RSEs.
    """

    def __init__(self, rse_ids: Optional['Iterable[str]'] = None, rse_data_cls: type[T] = RseData):
        self._rse_data_cls = rse_data_cls
        self.rse_id_to_data_map: dict[str, T] = {}
        if rse_ids is not None:
            for rse_id in rse_ids:
                self.rse_id_to_data_map[rse_id] = self._rse_data_cls(rse_id)

    def __getitem__(
            self,
            item: str
    ) -> "T":
        return self.get_or_create(item)

    def __setitem__(
            self,
            key: str,
            value: "T"
    ) -> None:
        rse_id = key
        rse_data = value
        self.rse_id_to_data_map[rse_id] = rse_data

    def __contains__(
            self,
            item: Any
    ) -> bool:
        if isinstance(item, RseData):
            return item.id in self.rse_id_to_data_map
        if isinstance(item, str):
            return item in self.rse_id_to_data_map
        return False

    def get(self, rse_id: str) -> "Optional[T]":
        return self.rse_id_to_data_map.get(rse_id)

    def get_or_create(self, rse_id: str) -> "T":
        rse_data = self.rse_id_to_data_map.get(rse_id)
        if rse_data is None:
            self.rse_id_to_data_map[rse_id] = rse_data = self._rse_data_cls(rse_id)
        return rse_data

    @transactional_session
    def ensure_loaded(
            self,
            rse_ids: Optional["Iterable[str]"] = None,
            load_name: bool = False,
            load_columns: bool = False,
            load_attributes: bool = False,
            load_info: bool = False,
            load_usage: bool = False,
            load_limits: bool = False,
            include_deleted: bool = False,
            *,
            session: "Session",
    ) -> None:
        RseData.bulk_load(
            rse_id_to_data={rse_id: self.get_or_create(rse_id) for rse_id in rse_ids} if rse_ids else self.rse_id_to_data_map,
            load_name=load_name,
            load_columns=load_columns,
            load_attributes=load_attributes,
            load_info=load_info,
            load_usage=load_usage,
            load_limits=load_limits,
            include_deleted=include_deleted,
            session=session,
        )


@stream_session
def _group_query_result_by_rse_id(
    stmt: "Executable",
    *,
    session: "Session"
) -> 'Iterator[tuple[str, list[Any]]]':
    """
    Given a sqlalchemy query statement which fetches rows of two elements: (rse_id, object) ordered by rse_id.
    Will execute the query and return objects grouped by rse_id: (rse_id, [object1, object2])
    """

    current_rse_id = None
    objects = []
    for rse_id, obj in session.execute(stmt):
        if current_rse_id != rse_id:
            if current_rse_id is not None:
                yield str(current_rse_id), objects

            current_rse_id = rse_id
            objects = []

        if obj is not None:
            objects.append(obj)

    if current_rse_id is not None:
        yield str(current_rse_id), objects


@transactional_session
def add_rse(
    rse: str,
    vo: str = 'def',
    deterministic: bool = True,
    volatile: bool = False,
    city: Optional[str] = None,
    region_code: Optional[str] = None,
    country_name: Optional[str] = None,
    continent: Optional[str] = None,
    time_zone: Optional[str] = None,
    ISP: Optional[str] = None,  # noqa: N803
    staging_area: bool = False,
    rse_type: RSEType = RSEType.DISK,
    longitude: Optional[float] = None,
    latitude: Optional[float] = None,
    ASN: Optional[str] = None,  # noqa: N803
    availability_read: Optional[bool] = None,
    availability_write: Optional[bool] = None,
    availability_delete: Optional[bool] = None,
    *,
    session: "Session"
) -> str:
    """
    Add a rse with the given location name.

    :param rse: the name of the new rse.
    :param vo: the vo to add the RSE to.
    :param deterministic: Boolean to know if the pfn is generated deterministically.
    :param volatile: Boolean for RSE cache.
    :param city: City for the RSE. Accessed by `locals()`.
    :param region_code: The region code for the RSE. Accessed by `locals()`.
    :param country_name: The country. Accessed by `locals()`.
    :param continent: The continent. Accessed by `locals()`.
    :param time_zone: Timezone. Accessed by `locals()`.
    :param ISP: Internet service provider. Accessed by `locals()`.
    :param staging_area: Staging area.
    :param rse_type: RSE type.
    :param latitude: Latitude coordinate of RSE.
    :param longitude: Longitude coordinate of RSE.
    :param ASN: Access service network. Accessed by `locals()`.
    :param availability_read: If the RSE is readable.
    :param availability_write: If the RSE is writable.
    :param availability_delete: If the RSE is deletable.
    :param session: The database session in use.
    """
    if isinstance(rse_type, str):
        rse_type = RSEType(rse_type)

    availability = Availability(availability_read, availability_write, availability_delete).integer
    new_rse = models.RSE(rse=rse, vo=vo, deterministic=deterministic, volatile=volatile,
                         staging_area=staging_area, rse_type=rse_type, longitude=longitude,
                         latitude=latitude, availability=availability, availability_read=availability_read,
                         availability_write=availability_write, availability_delete=availability_delete,

                         # The following fields will be deprecated, they are RSE attributes now.
                         # (Still in the code for backwards compatibility)
                         city=city, region_code=region_code, country_name=country_name,
                         continent=continent, time_zone=time_zone, ISP=ISP, ASN=ASN)
    try:
        new_rse.save(session=session)
    except IntegrityError:
        raise exception.Duplicate(f"RSE '{rse}' already exists!")
    except DatabaseError as error:
        raise exception.RucioException(error.args)

    # Add rse name as a RSE-Tag
    add_rse_attribute(rse_id=new_rse.id, key=rse, value=True, session=session)

    for setting in RSE_SETTINGS:
        # The value accessed by locals is defined in the code and it can not be
        # changed by a user request. This thus does not provide a scurity risk.
        setting_value = locals().get(setting, None)
        if setting_value:
            add_rse_attribute(rse_id=new_rse.id, key=setting, value=setting_value, session=session)

    # Add counter to monitor the space usage
    add_counter(rse_id=new_rse.id, session=session)

    return new_rse.id


@read_session
def rse_exists(
    rse: str,
    vo: str = 'def',
    include_deleted: bool = False,
    *,
    session: "Session"
) -> bool:
    """
    Checks to see if RSE exists.

    :param rse: Name of the rse.
    :param vo: The VO for the RSE.
    :param session: The database session in use.

    :returns: True if found, otherwise false.
    """
    stmt = select(
        models.RSE
    ).where(
        and_(models.RSE.rse == rse,
             models.RSE.vo == vo)
    )
    if not include_deleted:
        stmt = stmt.where(models.RSE.deleted == false())
    return True if session.execute(stmt).scalar() else False


@transactional_session
def del_rse(
    rse_id: str,
    *,
    session: "Session"
) -> None:
    """
    Disable a rse with the given rse id.

    :param rse_id: the rse id.
    :param session: The database session in use.
    """

    try:
        stmt = select(
            models.RSE
        ).where(
            and_(models.RSE.id == rse_id,
                 models.RSE.deleted == false())
        )
        db_rse = session.execute(stmt).scalar_one()
        rse_name = db_rse.rse
        if not rse_is_empty(rse_id=rse_id, session=session):
            raise exception.RSEOperationNotSupported('RSE \'%s\' is not empty' % rse_name)
    except sqlalchemy.orm.exc.NoResultFound:
        raise exception.RSENotFound('RSE with id \'%s\' cannot be found' % rse_id)
    db_rse.delete(session=session)
    try:
        del_rse_attribute(rse_id=rse_id, key=rse_name, session=session)
    except exception.RSEAttributeNotFound:
        pass


@transactional_session
def restore_rse(
    rse_id: str,
    *,
    session: "Session"
) -> None:
    """
    Restore a rse with the given rse id.

    :param rse_id: the rse id.
    :param session: The database session in use.
    """

    try:
        stmt = select(
            models.RSE
        ).where(
            and_(models.RSE.id == rse_id,
                 models.RSE.deleted == true())
        )
        db_rse = session.execute(stmt).scalar_one()
    except sqlalchemy.orm.exc.NoResultFound:
        raise exception.RSENotFound('RSE with id \'%s\' cannot be found' % rse_id)
    db_rse.deleted = False
    db_rse.deleted_at = None
    db_rse.save(session=session)
    rse_name = db_rse.rse
    add_rse_attribute(rse_id=rse_id, key=rse_name, value=True, session=session)


@read_session
def rse_is_empty(
    rse_id: str,
    *,
    session: "Session"
) -> bool:
    """
    Check if a RSE is empty.

    :param rse_id: the rse id.
    :param session: the database session in use.
    """

    is_empty = False
    try:
        is_empty = get_counter(rse_id, session=session)['bytes'] == 0
    except exception.CounterNotFound:
        is_empty = True
    return is_empty


@read_session
def _format_get_rse(
        db_rse: models.RSE,
        rse_attributes: Optional[dict[str, Any]] = None,
        *,
        session: "Session"
) -> dict[str, Any]:
    """
    Given a models.RSE object, return it formatted as expected by callers of get_rse
    """
    result = db_rse.to_dict()
    result['type'] = db_rse.rse_type
    if rse_attributes is not None:
        rse_settings = {key: rse_attributes[key] for key in RSE_SETTINGS if key in rse_attributes}
    else:
        stmt = select(
            models.RSEAttrAssociation
        ).where(
            and_(models.RSEAttrAssociation.rse_id == db_rse.id,
                 models.RSEAttrAssociation.key.in_(RSE_SETTINGS)),
        )
        rse_settings = {str(row.key): row.value for row in session.execute(stmt).scalars()}
    result.update(rse_settings)
    return result


@read_session
def get_rse(
    rse_id: str,
    *,
    session: "Session"
) -> dict[str, Any]:
    """
    Get a RSE or raise if it does not exist.

    :param rse_id:  The rse id.
    :param session: The database session in use.

    :raises RSENotFound: If referred RSE was not found in the database.
    """

    try:
        stmt = select(
            models.RSE
        ).where(
            and_(models.RSE.deleted == false(),
                 models.RSE.id == rse_id)
        )
        return _format_get_rse(session.execute(stmt).scalar_one(), session=session)
    except sqlalchemy.orm.exc.NoResultFound:
        raise exception.RSENotFound('RSE with id \'%s\' cannot be found' % rse_id)


@read_session
def get_rse_id(
    rse: str,
    vo: str = 'def',
    include_deleted: bool = True,
    *,
    session: "Session"
) -> str:
    """
    Get a RSE ID or raise if it does not exist.

    :param rse: the rse name.
    :param session: The database session in use.
    :param include_deleted: Flag to toggle finding rse's marked as deleted.

    :returns: The rse id.

    :raises RSENotFound: If referred RSE was not found in the database.
    """

    if include_deleted:
        if vo != 'def':
            cache_key = 'rse-id_{}@{}'.format(rse, vo).replace(' ', '.')
        else:
            cache_key = 'rse-id_{}'.format(rse).replace(' ', '.')
        result = REGION.get(cache_key)
        if not isinstance(result, NoValue):
            return result

    try:
        stmt = select(
            models.RSE.id
        ).where(
            and_(models.RSE.rse == rse,
                 models.RSE.vo == vo)
        )
        if not include_deleted:
            stmt = stmt.where(models.RSE.deleted == false())
        result = session.execute(stmt).scalar_one()
    except sqlalchemy.orm.exc.NoResultFound:
        raise exception.RSENotFound("RSE '%s' cannot be found in vo '%s'" % (rse, vo))

    if include_deleted:
        REGION.set(cache_key, result)
    return result


@read_session
def _get_rse_db_column(
    rse_id: str,
    column: "_ColumnsClauseArgument",
    cache_prefix: str,
    include_deleted: bool = True,
    *,
    session: "Session"
) -> Any:
    if include_deleted:
        cache_key = '{}_{}'.format(cache_prefix, rse_id)
        result = REGION.get(cache_key)
        if not isinstance(result, NoValue):
            return result

    try:
        stmt = select(
            column
        ).where(
            models.RSE.id == rse_id
        )
        if not include_deleted:
            stmt = stmt.where(models.RSE.deleted == false())
        result = session.execute(stmt).scalar_one()
    except sqlalchemy.orm.exc.NoResultFound:
        raise exception.RSENotFound('RSE with ID \'%s\' cannot be found' % rse_id)

    if include_deleted:
        REGION.set(cache_key, result)
    return result


@read_session
def get_rse_name(
    rse_id: str,
    include_deleted: bool = True,
    *,
    session: "Session"
) -> str:
    """
    Get a RSE name or raise if it does not exist.

    :param rse_id: the rse uuid from the database.
    :param session: The database session in use.
    :param include_deleted: Flag to toggle finding rse's marked as deleted.

    :returns: The rse name.

    :raises RSENotFound: If referred RSE was not found in the database.
    """
    return _get_rse_db_column(
        rse_id=rse_id,
        column=models.RSE.rse,
        cache_prefix='rse-name',
        include_deleted=include_deleted,
        session=session
    )


@read_session
def get_rse_vo(
    rse_id: str,
    include_deleted: bool = True,
    *,
    session: "Session"
) -> str:
    """
    Get the VO for a given RSE id.

    :param rse_id: the rse uuid from the database.
    :param session: the database session in use.
    :param include_deleted: Flag to toggle finding rse's marked as deleted.

    :returns The vo name.

    :raises RSENotFound: If referred RSE was not found in database.
    """
    return _get_rse_db_column(
        rse_id=rse_id,
        column=models.RSE.vo,
        cache_prefix='rse-vo',
        include_deleted=include_deleted,
        session=session
    )


@read_session
def list_rses(filters: Optional[dict[str, Any]] = None, *, session: "Session") -> list[dict[str, Any]]:
    """
    Returns a list of all RSEs.

    :param filters: dictionary of attributes by which the results should be filtered.
    :param session: The database session in use.

    :returns: a list of dictionaries.
    """

    filters = filters or {}
    filters = filters.copy()  # Make a copy, so we can pop() without affecting the object `filters` outside this function

    stmt = select(
        models.RSE
    ).where(
        models.RSE.deleted == false()
    )
    if filters:
        if 'availability' in filters and ('availability_read' in filters or 'availability_write' in filters or 'availability_delete' in filters):
            raise exception.InvalidObject('Cannot use availability and read, write, delete filter at the same time.')

        if 'availability' in filters:
            availability = Availability.from_integer(filters['availability'])
            filters['availability_read'] = availability.read
            filters['availability_write'] = availability.write
            filters['availability_delete'] = availability.delete
            del filters['availability']

        for (k, v) in filters.items():
            if hasattr(models.RSE, k):
                if k == 'rse_type':
                    stmt = stmt.where(getattr(models.RSE, k) == RSEType[v])
                else:
                    stmt = stmt.where(getattr(models.RSE, k) == v)
            else:
                attr_assoc_alias = aliased(models.RSEAttrAssociation)
                stmt = stmt.join(
                    attr_assoc_alias,
                    and_(attr_assoc_alias.rse_id == models.RSE.id,
                         attr_assoc_alias.key == k,
                         attr_assoc_alias.value == v)
                )
    stmt = stmt.order_by(
        models.RSE.rse
    )

    return [row.to_dict() for row in session.execute(stmt).scalars()]


@transactional_session
def add_rse_attribute(
    rse_id: str,
    key: str,
    value: Union[bool, str],
    *,
    session: "Session"
) -> bool:
    """ Adds a RSE attribute.

    :param rse_id: the rse id.
    :param key: the key name.
    :param value: the value name.
    :param issuer: The issuer account.
    :param session: The database session in use.

    :returns: True is successful
    """
    try:
        new_rse_attr = models.RSEAttrAssociation(rse_id=rse_id, key=key, value=value)
        new_rse_attr = session.merge(new_rse_attr)
        new_rse_attr.save(session=session)
    except IntegrityError:
        rse = get_rse_name(rse_id=rse_id, session=session)
        raise exception.Duplicate(f"RSE attribute '{key}-{value}' for RSE '{rse}' already exists!")
    return True


@transactional_session
def del_rse_attribute(
    rse_id: str,
    key: str,
    *,
    session: "Session"
) -> bool:
    """
    Delete a RSE attribute.

    :param rse_id: the id of the rse.
    :param key: the attribute key.
    :param session: The database session in use.

    :return: True if RSE attribute was deleted.
    """
    try:
        stmt = select(
            models.RSEAttrAssociation
        ).where(
            and_(models.RSEAttrAssociation.rse_id == rse_id,
                 models.RSEAttrAssociation.key == key)
        )
        rse_attr = session.execute(stmt).scalar_one()
    except sqlalchemy.orm.exc.NoResultFound:
        raise exception.RSEAttributeNotFound('RSE attribute \'%s\' cannot be found' % key)
    rse_attr.delete(session=session)
    return True


@read_session
def list_rse_attributes(
    rse_id: str,
    use_cache: bool = False,
    *,
    session: "Session"
) -> dict[str, Union[str, bool]]:
    """
    List RSE attributes for a RSE.

    :param rse_id:  The RSE id.
    :param use_cache: decides if cache will be used or not
    :param session: The database session in use.

    :returns: A dictionary with RSE attributes for a RSE.
    """
    cache_key = 'rse_attributes_%s' % rse_id
    if use_cache:
        value = REGION.get(cache_key)

        if not isinstance(value, NoValue):
            return value

    rse_attrs = {}

    stmt = select(
        models.RSEAttrAssociation
    ).where(
        models.RSEAttrAssociation.rse_id == rse_id
    )
    for attr in session.execute(stmt).scalars():
        rse_attrs[attr.key] = attr.value

    if use_cache:
        REGION.set(cache_key, rse_attrs)

    return rse_attrs


@stream_session
def _fetch_many_rses_attributes(
        rse_id_temp_table: Any,
        keys: Optional['Iterable[str]'] = None,
        *,
        session: "Session"
) -> 'Iterator[tuple[str, dict[str, Any]]]':
    """
    Given a temporary table pre-filled with RSE IDs, fetch the attributes of these RSEs.
    It's possible to only fetch a subset of attributes by setting the `keys` parameter.
    """

    stmt = select(
        rse_id_temp_table.id,
        models.RSEAttrAssociation,
    ).outerjoin_from(
        rse_id_temp_table,
        models.RSEAttrAssociation,
        models.RSEAttrAssociation.rse_id == rse_id_temp_table.id
    ).order_by(
        rse_id_temp_table.id,
    )

    if keys:
        stmt = stmt.where(
            models.RSEAttrAssociation.key.in_(keys)
        )

    for rse_id, attribute_list in _group_query_result_by_rse_id(stmt, session=session):
        yield rse_id, {attr.key: attr.value for attr in attribute_list}


@read_session
def has_rse_attribute(
    rse_id: str,
    key: str,
    *,
    session: "Session"
) -> bool:
    """
    Indicates whether the named key is present for the RSE.

    :param rse_id: The RSE id.
    :param key: The key for the attribute.
    :param session: The database session in use.

    :returns: True or False
    """
    stmt = select(
        models.RSEAttrAssociation.value
    ).where(
        and_(models.RSEAttrAssociation.rse_id == rse_id,
             models.RSEAttrAssociation.key == key)
    )
    if session.execute(stmt).scalar():
        return True
    return False


@read_session
def get_rses_with_attribute(
    key: str,
    *,
    session: "Session"
) -> list[dict[str, Any]]:
    """
    Return all RSEs with a certain attribute.

    :param key: The key for the attribute.
    :param session: The database session in use.

    :returns: List of rse dictionaries
    """
    rse_list = []

    stmt = select(
        models.RSE
    ).where(
        models.RSE.deleted == false()
    ).join(
        models.RSEAttrAssociation,
        and_(models.RSEAttrAssociation.rse_id == models.RSE.id,
             models.RSEAttrAssociation.key == key)
    )

    for db_rse in session.execute(stmt).scalars():
        rse_list.append(db_rse.to_dict())

    return rse_list


@read_session
def get_rses_with_attribute_value(
    key: str,
    value: Union[bool, str],
    vo: str = 'def',
    *,
    session: "Session"
) -> list[dict[str, str]]:
    """
    Return all RSEs with a certain attribute.

    :param key: The key for the attribute.
    :param value: The value for the attribute.
    :param session: The database session in use.

    :returns: List of rse dictionaries with the rse_id and rse_name
    """
    if vo != 'def':
        cache_key = 'av-%s-%s@%s' % (key, value, vo)
    else:
        cache_key = 'av-%s-%s' % (key, value)

    result = REGION.get(cache_key)
    if isinstance(result, NoValue):

        rse_list = []

        stmt = select(
            models.RSE.id,
            models.RSE.rse,
        ).where(
            and_(models.RSE.deleted == false(),
                 models.RSE.vo == vo)
        ).join(
            models.RSEAttrAssociation,
            and_(models.RSEAttrAssociation.rse_id == models.RSE.id,
                 models.RSEAttrAssociation.key == key,
                 models.RSEAttrAssociation.value == value)
        )

        for row in session.execute(stmt):
            rse_list.append({
                'rse_id': row.id,
                'rse_name': row.rse
            })

        REGION.set(cache_key, rse_list)
        return rse_list

    return result


@overload
def get_rse_attribute(rse_id: str, key: Literal['sign_url'], use_cache: bool = True, *, session: "Session") -> Optional[SUPPORTED_SIGN_URL_SERVICES_LITERAL]:
    ...


@overload
def get_rse_attribute(rse_id: str, key: Literal['sign_url'], use_cache: bool = True) -> Optional[SUPPORTED_SIGN_URL_SERVICES_LITERAL]:
    ...


@overload
def get_rse_attribute(rse_id: str, key: 'RSE_ATTRS_STR', use_cache: bool = True) -> Optional[str]:
    ...


@overload
def get_rse_attribute(rse_id: str, key: 'RSE_ATTRS_STR', use_cache: bool = True, *, session: "Session") -> Optional[str]:
    ...


@overload
def get_rse_attribute(rse_id: str, key: 'RSE_ATTRS_BOOL', use_cache: bool = True) -> Optional[bool]:
    ...


@overload
def get_rse_attribute(rse_id: str, key: 'RSE_ATTRS_BOOL', use_cache: bool = True, *, session: "Session") -> Optional[bool]:
    ...


@read_session
def get_rse_attribute(rse_id: str, key: str, use_cache: bool = True, *, session: "Session") -> Optional[Union[str, bool]]:
    """
    Retrieve RSE attribute value. If it is not cached, look it up in the
    database. If the value exists and is not cached, it will be added to the
    cache.

    :param rse_id: The RSE id.
    :param key: The key for the attribute.
    :param session: The database session in use.

    :returns: The value for the rse attribute, None if it does not exist.
    """
    cache_key = f'rse_attributes_{rse_id}_{key}'
    if use_cache:
        value = REGION.get(cache_key)

        if not isinstance(value, NoValue):
            return value

    stmt = select(
        models.RSEAttrAssociation.value
    ).where(
        and_(models.RSEAttrAssociation.rse_id == rse_id,
             models.RSEAttrAssociation.key == key)
    )
    value = session.execute(stmt).scalar_one_or_none()

    if use_cache:
        REGION.set(cache_key, value)

    return value


def get_rse_supported_checksums_from_attributes(rse_attributes: dict[str, Any]) -> list[str]:
    """
    Parse the RSE attribute defining the checksum supported by the RSE
    :param rse_attributes: attributes retrieved using list_rse_attributes
    :returns: A list of the names of supported checksums indicated by the specified attributes.
    """
    return parse_checksum_support_attribute(rse_attributes.get(CHECKSUM_KEY, ''))


def parse_checksum_support_attribute(checksum_attribute: str) -> list[str]:
    """
    Parse the checksum support RSE attribute.
    :param checksum_attribute: The value of the RSE attribute storing the checksum value

    :returns: The list of checksums supported by the selected RSE.
              If the list is empty (aka attribute is not set) it returns all the default checksums.
              Use 'none' to explicitly tell the RSE does not support any checksum algorithm.
    """

    if not checksum_attribute:
        return GLOBALLY_SUPPORTED_CHECKSUMS

    supported_checksum_list = [c.strip() for c in checksum_attribute.split(',') if c.strip()]

    if 'none' in supported_checksum_list:
        return []
    else:
        return supported_checksum_list


@transactional_session
def set_rse_usage(
    rse_id: str,
    source: str,
    used: int,
    free: int,
    files: Optional[int] = None,
    *,
    session: "Session"
) -> bool:
    """
    Set RSE usage information.

    :param rse_id: the location id.
    :param source: The information source, e.g. srm.
    :param used: the used space in bytes.
    :param free: the free in bytes.
    :param files: the number of files
    :param session: The database session in use.

    :returns: True if successful, otherwise false.
    """
    rse_usage = models.RSEUsage(rse_id=rse_id, source=source, used=used, free=free, files=files)
    # versioned_session(session)
    rse_usage = session.merge(rse_usage)
    rse_usage.save(session=session)

    # rse_usage_history = models.RSEUsage.__history_mapper__.class_(rse_id=rse.id, source=source, used=used, free=free)
    # rse_usage_history.save(session=session)

    return True


@read_session
def get_rse_usage(
    rse_id: str,
    source: Optional[str] = None,
    per_account: bool = False,
    *,
    session: "Session"
) -> list[dict[str, Any]]:
    """
    get rse usage information.

    :param rse_id:  The RSE id.
    :param source: The information source, e.g. srm.
    :param session: The database session in use.
    :param per_account: Boolean whether the usage should be also calculated per account or not.

    :returns: List of RSE usage data.
    """

    stmt_rse_usage = select(
        models.RSEUsage
    ).where(
        models.RSEUsage.rse_id == rse_id
    )

    if source:
        stmt_rse_usage = stmt_rse_usage.where(
            models.RSEUsage.source == source
        )

    db_usages = session.execute(stmt_rse_usage).scalars()
    return _format_get_rse_usage(rse_id=rse_id, db_usages=db_usages, per_account=per_account, session=session)


def _format_get_rse_usage(
        rse_id: str,
        db_usages: 'Iterable[models.RSEUsage]',
        per_account: bool,
        *,
        session: "Session"
) -> list[dict[str, Any]]:

    usage = list()
    for db_usage in db_usages:
        total = (db_usage.free or 0) + (db_usage.used or 0)
        rse_usage = {'rse_id': rse_id,
                     'source': db_usage.source,
                     'used': db_usage.used,
                     'free': db_usage.free,
                     'total': total,
                     'files': db_usage.files,
                     'updated_at': db_usage.updated_at}
        if per_account and db_usage.source == 'rucio':
            stmt_account_usage = select(
                models.AccountUsage
            ).where(
                models.AccountUsage.rse_id == rse_id
            )
            account_usages = []
            for row in session.execute(stmt_account_usage).scalars():
                if row.bytes != 0:
                    percentage = round(float(row.bytes) / float(total) * 100, 2) if total else 0
                    account_usages.append({'used': row.bytes, 'account': row.account, 'percentage': percentage})
            account_usages.sort(key=lambda x: x['used'], reverse=True)
            rse_usage['account_usages'] = account_usages
        usage.append(rse_usage)
    return usage


@transactional_session
def set_rse_limits(rse_id: str, name: str, value: int, *, session: 'Session') -> bool:
    """
    Set RSE limits.

    :param rse_id: The RSE id.
    :param name: The name of the limit.
    :param value: The feature value.
    :param session: The database session in use.

    :returns: True if successful, otherwise false.
    """
    rse_limit = models.RSELimit(rse_id=rse_id, name=name, value=value)
    rse_limit = session.merge(rse_limit)
    rse_limit.save(session=session)
    return True


@read_session
def get_rse_limits(rse_id: str, name: Optional[str] = None, *, session: 'Session') -> dict[str, int]:
    """
    Get RSE limits.

    :param rse_id: The RSE id.
    :param name: A Limit name.

    :returns: A dictionary with the limits {'limit.name': limit.value}.
    """

    stmt = select(
        models.RSELimit
    ).where(
        models.RSELimit.rse_id == rse_id
    )
    if name:
        stmt = stmt.where(
            models.RSELimit.name == name
        )
    return {limit.name: limit.value for limit in session.execute(stmt).scalars()}


@transactional_session
def delete_rse_limits(rse_id: str, name: "Optional[str]" = None, *, session: 'Session') -> None:
    """
    Delete RSE limit.

    :param rse_id: The RSE id.
    :param name: The name of the limit.
    """
    try:
        stmt = delete(
            models.RSELimit
        ).where(
            models.RSELimit.rse_id == rse_id,
        )
        if name is not None:
            stmt = stmt.where(
                models.RSELimit.name == name
            )
        session.execute(stmt)
    except IntegrityError as error:
        raise exception.RucioException(error.args)


def _sanitize_rse_transfer_limit_dict(limit_dict: dict[str, Any]) -> dict[str, Any]:
    if limit_dict['activity'] == 'all_activities':
        limit_dict['activity'] = None
    return limit_dict


@read_session
def get_rse_transfer_limits(
    rse_id: str,
    activity: Optional[str] = None,
    *,
    session: "Session"
) -> dict[str, Any]:
    """
    Get RSE transfer limits.

    :param rse_id: The RSE id.
    :param activity: The activity.

    :returns: A dictionary with the limits {'limit.direction': {'limit.activity': limit}}.
    """
    try:
        stmt = select(
            models.TransferLimit
        ).join_from(
            models.RSETransferLimit,
            models.TransferLimit,
            and_(models.RSETransferLimit.limit_id == models.TransferLimit.id,
                 models.RSETransferLimit.rse_id == rse_id)
        )
        if activity:
            stmt = stmt.where(
                or_(models.TransferLimit.activity == activity,
                    models.TransferLimit.activity == 'all_activities')
            )

        limits = {}
        for limit in session.execute(stmt).scalars():
            limit_dict = _sanitize_rse_transfer_limit_dict(limit.to_dict())
            limits.setdefault(limit_dict['direction'], {}).setdefault(limit_dict['activity'], limit_dict)

        return limits
    except IntegrityError as error:
        raise exception.RucioException(error.args)


@stream_session
def list_rse_usage_history(
    rse_id: str,
    source: Optional[str] = None,
    *,
    session: "Session"
) -> 'Iterator[dict[str, Any]]':
    """
    List RSE usage history information.

    :param rse_id: The RSE id.
    :param source: The source of the usage information (srm, rucio).
    :param session: The database session in use.

    :returns: A list of historic RSE usage.
    """
    stmt = select(
        models.RSEUsageHistory
    ).where(
        models.RSEUsageHistory.rse_id == rse_id
    ).order_by(
        desc(models.RSEUsageHistory.updated_at)
    )
    if source:
        stmt = stmt.where(
            models.RSEUsageHistory.source == source
        )

    rse = get_rse_name(rse_id=rse_id, session=session)
    for usage in session.execute(stmt).yield_per(5).scalars():
        yield ({'rse_id': rse_id,
                'rse': rse,
                'source': usage.source,
                'used': usage.used if usage.used else 0,
                'total': usage.used if usage.used else 0 + usage.free if usage.free else 0,
                'free': usage.free if usage.free else 0,
                'updated_at': usage.updated_at})


@transactional_session
def add_protocol(
    rse_id: str,
    parameter: dict[str, Any],
    *,
    session: "Session"
) -> models.RSEProtocol:
    """
    Add a protocol to an existing RSE.

    :param rse_id: the ID of the new RSE.
    :param parameter: parameters of the new protocol entry.
    :param session: The database session in use.

    :raises RSENotFound: If RSE is not found.
    :raises RSEOperationNotSupported: If no scheme supported the requested operation for the given RSE.
    :raises RSEProtocolDomainNotSupported: If an undefined domain was provided.
    :raises RSEProtocolPriorityError: If the provided priority for the scheme is to big or below zero.
    :raises Duplicate: If scheme with identifier, hostname and port already exists
                       for the given RSE.
    """

    rse = ""
    try:
        rse = get_rse_name(rse_id=rse_id, session=session, include_deleted=False)
    except exception.RSENotFound:
        raise exception.RSENotFound('RSE id \'%s\' not found' % rse_id)
    # Insert new protocol entry
    parameter['rse_id'] = rse_id

    # Default values
    parameter['port'] = parameter.get('port', 0)
    parameter['hostname'] = parameter.get('hostname', 'localhost')

    # Transform nested domains to match DB schema e.g. [domains][lan][read] => [read_lan]
    if 'domains' in parameter:
        for domain in parameter['domains']:
            if domain not in utils.rse_supported_protocol_domains():
                raise exception.RSEProtocolDomainNotSupported(f"The protocol domain '{domain}' is not defined in the schema.")
            for op in parameter['domains'][domain]:
                if op not in RSE_ALL_SUPPORTED_PROTOCOL_OPERATIONS:
                    raise exception.RSEOperationNotSupported(f"Operation '{op}' not defined in schema.")
                op_name = op if op.startswith('third_party_copy') else f'{op}_{domain}'.lower()
                priority = parameter['domains'][domain][op]
                if (type(priority) is not int or priority < 0) and priority is not None:
                    raise exception.RSEProtocolPriorityError(f"The provided priority ({priority}) for operation '{op}' in domain '{domain}' is not supported.")
                parameter[op_name] = priority
        del parameter['domains']

    if ('extended_attributes' in parameter) and parameter['extended_attributes']:
        try:
            parameter['extended_attributes'] = json.dumps(parameter['extended_attributes'], separators=(',', ':'))
        except ValueError:
            pass  # String is not JSON

    if parameter['scheme'] == 'srm':
        if ('extended_attributes' not in parameter) or ('web_service_path' not in parameter['extended_attributes']):
            raise exception.InvalidObject('Missing values! For SRM, extended_attributes and web_service_path must be specified')

    try:
        new_protocol = models.RSEProtocol()
        new_protocol.update(parameter)
        new_protocol.save(session=session)
    except (IntegrityError, FlushError, OperationalError) as error:
        if ('UNIQUE constraint failed' in error.args[0]) or ('conflicts with persistent instance' in error.args[0]) \
                or match('.*IntegrityError.*ORA-00001: unique constraint.*RSE_PROTOCOLS_PK.*violated.*', error.args[0]) \
                or match('.*IntegrityError.*1062.*Duplicate entry.*for key.*', error.args[0]) \
                or match('.*IntegrityError.*duplicate key value violates unique constraint.*', error.args[0]) \
                or match('.*UniqueViolation.*duplicate key value violates unique constraint.*', error.args[0]) \
                or match('.*IntegrityError.*columns.*are not unique.*', error.args[0]):
            raise exception.Duplicate('Protocol \'%s\' on port %s already registered for  \'%s\' with hostname \'%s\'.' % (parameter['scheme'], parameter['port'], rse, parameter['hostname']))
        elif 'may not be NULL' in error.args[0] \
                or match('.*IntegrityError.*ORA-01400: cannot insert NULL into.*RSE_PROTOCOLS.*IMPL.*', error.args[0]) \
                or match('.*IntegrityError.*Column.*cannot be null.*', error.args[0]) \
                or match('.*IntegrityError.*null value in column.*violates not-null constraint.*', error.args[0]) \
                or match('.*IntegrityError.*NOT NULL constraint failed.*', error.args[0]) \
                or match('.*NotNullViolation.*null value in column.*violates not-null constraint.*', error.args[0]) \
                or match('.*OperationalError.*cannot be null.*', error.args[0]):
            raise exception.InvalidObject('Missing values!')

        raise exception.RucioException(error.args)
    return new_protocol


@read_session
def get_rse_protocols(
    rse_id: str,
    schemes: Optional[list[str]] = None,
    *,
    session: "Session"
) -> types.RSESettingsDict:
    """
    Returns protocol information. Parameter combinations are: (operation OR default) XOR scheme.

    :param rse_id: The id of the rse.
    :param schemes: a list of schemes to filter by.
    :param session: The database session.

    :returns: A dict with RSE information and supported protocols

    :raises RSENotFound: If RSE is not found.
    """

    _rse = get_rse(rse_id=rse_id, session=session)
    if not _rse:
        raise exception.RSENotFound('RSE with id \'%s\' not found' % rse_id)

    terms = [models.RSEProtocol.rse_id == rse_id]
    if schemes:
        if not isinstance(schemes, list):
            schemes = [schemes]
        terms.extend([models.RSEProtocol.scheme.in_(schemes)])

    stmt = select(
        models.RSEProtocol
    ).where(
        *terms
    )

    _protocols = session.execute(stmt).scalars().all()
    return _format_get_rse_protocols(rse=_rse, db_protocols=_protocols, session=session)


def _format_get_rse_protocols(
        rse: "models.RSE | dict[str, Any]",
        db_protocols: 'Iterable[models.RSEProtocol]',
        rse_attributes: Optional[dict[str, Any]] = None,
        *,
        session: "Session"
) -> types.RSESettingsDict:
    _rse = rse
    if rse_attributes:
        lfn2pfn_algorithm = rse_attributes.get(RseAttr.LFN2PFN_ALGORITHM)
    else:
        lfn2pfn_algorithm = get_rse_attribute(_rse['id'], RseAttr.LFN2PFN_ALGORITHM, session=session)
    # Resolve LFN2PFN default algorithm as soon as possible.  This way, we can send back the actual
    # algorithm name in response to REST queries.
    if not lfn2pfn_algorithm:
        lfn2pfn_algorithm = get_lfn2pfn_algorithm_default()

    # Copy verify_checksum from the attributes, later: assume True if not specified
    if rse_attributes:
        verify_checksum = rse_attributes.get(RseAttr.VERIFY_CHECKSUM)
    else:
        verify_checksum = get_rse_attribute(_rse['id'], RseAttr.VERIFY_CHECKSUM, session=session)

    # Copy sign_url from the attributes
    if rse_attributes:
        sign_url = rse_attributes.get(RseAttr.SIGN_URL)
    else:
        sign_url = get_rse_attribute(_rse['id'], RseAttr.SIGN_URL, session=session)

    info = {'availability_delete': _rse['availability_delete'],
            'availability_read': _rse['availability_read'],
            'availability_write': _rse['availability_write'],
            'credentials': None,
            'deterministic': _rse['deterministic'],
            'domain': utils.rse_supported_protocol_domains(),
            'id': _rse['id'],
            'lfn2pfn_algorithm': lfn2pfn_algorithm,
            'protocols': list(),
            'qos_class': _rse['qos_class'],
            'rse': _rse['rse'],
            'rse_type': _rse['rse_type'].name,
            'sign_url': sign_url,
            'staging_area': _rse['staging_area'],
            'verify_checksum': verify_checksum if verify_checksum is not None else True,
            'volatile': _rse['volatile']}

    for op in RSE_ALL_SUPPORTED_PROTOCOL_OPERATIONS:
        info['%s_protocol' % op] = 1  # 1 indicates the default protocol

    for row in db_protocols:
        p = {'hostname': row.hostname,
             'scheme': row.scheme,
             'port': row.port,
             'prefix': row.prefix if row.prefix is not None else '',
             'impl': row.impl,
             'domains': {
                 'lan': {'read': row.read_lan,
                         'write': row.write_lan,
                         'delete': row.delete_lan},
                 'wan': {'read': row.read_wan,
                         'write': row.write_wan,
                         'delete': row.delete_wan,
                         'third_party_copy_read': row.third_party_copy_read,
                         'third_party_copy_write': row.third_party_copy_write}
             },
             'extended_attributes': row.extended_attributes}

        try:
            p['extended_attributes'] = json.load(StringIO(p['extended_attributes']))
        except ValueError:
            pass  # If value is not a JSON string

        info['protocols'].append(p)
    info['protocols'] = sorted(info['protocols'], key=lambda p: (p['hostname'], p['scheme'], p['port']))
    return info


@read_session
def get_rse_info(rse_id: str, *, session: "Session") -> types.RSESettingsDict:
    """
    For historical reasons, related to usage of rsemanager, "rse_info" is equivalent to
    a cached call to get_rse_protocols without any schemes set.

    :param rse_id: The id of the rse.
    :param session: The database session.
    :returns: A dict with RSE information and supported protocols
    """
    key = 'rse_info_%s' % rse_id
    result = REGION.get(key)
    if isinstance(result, NoValue):
        result = get_rse_protocols(rse_id=rse_id, session=session)
        REGION.set(key, result)
    return result


@transactional_session
def update_protocols(
    rse_id: str,
    scheme: str,
    data: dict[str, Any],
    hostname: str,
    port: int,
    *,
    session: "Session"
) -> None:
    """
    Update an existing protocol entry for an RSE.

    :param rse_id: the ID of the RSE.
    :param scheme: Protocol identifier.
    :param data: Dict with new values (keys must match column names in the database).
    :param hostname: Hostname defined for the scheme, used if more than one scheme
                     is registered with the same identifier.
    :param port: The port registered for the hostname, used if more than one scheme
                 is registered with the same identifier and hostname.
    :param session: The database session in use.

    :raises RSENotFound: If RSE is not found.
    :raises RSEProtocolNotSupported: If no matching protocol was found for the given RSE.
    :raises RSEOperationNotSupported: If no protocol supported the requested operation for the given RSE.
    :raises RSEProtocolDomainNotSupported: If an undefined domain was provided.
    :raises RSEProtocolPriorityError: If the provided priority for the protocol is too big or below zero.
    :raises KeyNotFound: Invalid data for update provided.
    :raises Duplicate: If protocol with identifier, hostname and port already exists
                       for the given RSE.
    """

    # Transform nested domains to match DB schema e.g. [domains][lan][read] => [read_lan]
    if 'domains' in data:
        for domain in data['domains']:
            if domain not in utils.rse_supported_protocol_domains():
                raise exception.RSEProtocolDomainNotSupported(f"The protocol domain '{domain}' is not defined in the schema.")
            for op in data['domains'][domain]:
                if op not in RSE_ALL_SUPPORTED_PROTOCOL_OPERATIONS:
                    raise exception.RSEOperationNotSupported(f"Operation '{op}' not defined in schema.")
                op_name = op if op.startswith('third_party_copy') else f'{op}_{domain}'.lower()
                priority = data['domains'][domain][op]
                if (type(priority) is not int or priority < 0) and priority is not None:
                    raise exception.RSEProtocolPriorityError(f"The provided priority ({priority}) for operation '{op}' in domain '{domain}' is not supported.")
                data[op_name] = priority
        del data['domains']

    if 'extended_attributes' in data:
        try:
            data['extended_attributes'] = json.dumps(data['extended_attributes'], separators=(',', ':'))
        except ValueError:
            pass  # String is not JSON

    try:
        rse = get_rse_name(rse_id=rse_id, session=session, include_deleted=False)
    except exception.RSENotFound:
        raise exception.RSENotFound('RSE with id \'%s\' not found' % rse_id)

    terms = [models.RSEProtocol.rse_id == rse_id,
             models.RSEProtocol.scheme == scheme,
             models.RSEProtocol.hostname == hostname,
             models.RSEProtocol.port == port]

    try:
        stmt = select(
            models.RSEProtocol
        ).where(
            *terms
        )
        up = session.execute(stmt).scalar()
        if up is None:
            msg = 'RSE \'%s\' does not support protocol \'%s\' for hostname \'%s\' on port \'%s\'' % (rse, scheme, hostname, port)
            raise exception.RSEProtocolNotSupported(msg)
        up.update(data, flush=True, session=session)
    except (IntegrityError, OperationalError) as error:
        if 'UNIQUE'.lower() in error.args[0].lower() or 'Duplicate' in error.args[0]:  # Covers SQLite, Oracle and MySQL error
            raise exception.Duplicate('Protocol \'%s\' on port %s already registered for  \'%s\' with hostname \'%s\'.' % (scheme, port, rse, hostname))
        elif 'may not be NULL' in error.args[0] or "cannot be null" in error.args[0]:
            raise exception.InvalidObject('Missing values: %s' % error.args[0])
        raise error
    except DatabaseError as error:
        if match('.*DatabaseError.*ORA-01407: cannot update .*RSE_PROTOCOLS.*IMPL.*to NULL.*', error.args[0]):
            raise exception.InvalidObject('Invalid values !')
        raise error


@transactional_session
def del_protocols(
    rse_id: str,
    scheme: str,
    hostname: Optional[str] = None,
    port: Optional[int] = None,
    *,
    session: "Session"
) -> None:
    """
    Delete one or more existing protocol entries for an RSE.

    :param rse_id: the ID of the RSE.
    :param scheme: Protocol identifier.
    :param hostname: Hostname defined for the scheme, used if more than one scheme
                     is registered with the same identifier.
    :param port: The port registered for the hostname, used if more than one scheme
                     is registered with the same identifier and hostname.
    :param session: The database session in use.

    :raises RSENotFound: If RSE is not found.
    :raises RSEProtocolNotSupported: If no matching scheme was found for the given RSE.
    """
    try:
        rse_name = get_rse_name(rse_id=rse_id, session=session, include_deleted=False)
    except exception.RSENotFound:
        raise exception.RSENotFound('RSE \'%s\' not found' % rse_id)
    terms = [models.RSEProtocol.rse_id == rse_id, models.RSEProtocol.scheme == scheme]
    if hostname is not None:
        terms.append(models.RSEProtocol.hostname == hostname)
        if port is not None:
            terms.append(models.RSEProtocol.port == port)
    stmt = select(
        models.RSEProtocol
    ).where(
        *terms
    )
    p = session.execute(stmt).scalars().all()

    if not p:
        msg = 'RSE \'%s\' does not support protocol \'%s\'' % (rse_name, scheme)
        msg += ' for hostname \'%s\'' % hostname if hostname else ''
        msg += ' on port \'%s\'' % port if port else ''
        raise exception.RSEProtocolNotSupported(msg)

    for row in p:
        row.delete(session=session)


MUTABLE_RSE_PROPERTIES = {
    'name',
    'availability_read',
    'availability_write',
    'availability_delete',
    'latitude',
    'longitude',
    'time_zone',
    'rse_type',
    'volatile',
    'deterministic',
    'region_code',
    'country_name',
    'city',
    'staging_area',
    'qos_class',
    'continent',
    'availability'
}


@transactional_session
def update_rse(rse_id: str, parameters: dict[str, Any], *, session: "Session"):
    """
    Update RSE properties like availability or name.

    :param rse_id: the id of the new rse.
    :param parameters: A dictionary with property (name, read, write, delete as keys).
    :param session: The database session in use.

    :raises RSENotFound: If RSE is not found.
    :raises InputValidationError: If a parameter does not exist. Nothing will be added then.
    """
    for key in parameters.keys():
        if key not in MUTABLE_RSE_PROPERTIES:
            raise exception.InputValidationError(f"The key '{key}' does not exist for RSE properties.")

    try:
        stmt = select(
            models.RSE
        ).where(
            models.RSE.id == rse_id
        )
        db_rse = session.execute(stmt).scalar_one()
    except sqlalchemy.orm.exc.NoResultFound:
        raise exception.RSENotFound('RSE with ID \'%s\' cannot be found' % rse_id)
    old_rse_name = db_rse.rse

    param = {}

    if 'availability' in parameters:
        availability = Availability.from_integer(parameters['availability'])
        param['availability_read'] = availability.read
        param['availability_write'] = availability.write
        param['availability_delete'] = availability.delete

    for key in parameters:
        if key == 'name' and parameters['name'] != old_rse_name:  # Needed due to wrongly setting name in pre1.22.7 clients
            param['rse'] = parameters['name']
        elif key in MUTABLE_RSE_PROPERTIES - {'name'}:
            param[key] = parameters[key]

    # handle null-able keys
    for key in parameters:
        if key in ['qos_class']:
            if param[key] and param[key].lower() in ['', 'none', 'null']:
                param[key] = None

    # handle rse settings
    for setting in set(param.keys()).intersection(RSE_SETTINGS):
        if has_rse_attribute(rse_id, setting, session=session):
            del_rse_attribute(rse_id, setting, session=session)
        add_rse_attribute(rse_id, setting, param[setting], session=session)

    db_rse.update(param, session=session)
    if 'rse' in param:
        add_rse_attribute(rse_id=rse_id, key=parameters['name'], value=True, session=session)
        del_rse_attribute(rse_id=rse_id, key=old_rse_name, session=session)


@read_session
def export_rse(
    rse_id: str,
    *,
    session: "Session"
) -> dict[str, Any]:
    """
    Get the internal representation of an RSE.

    :param rse_id: The RSE id.

    :returns: A dictionary with the internal representation of an RSE.
    """

    stmt = select(
        models.RSE
    ).where(
        models.RSE.id == rse_id
    )

    rse_data = {}
    for _rse in session.execute(stmt).scalars():
        for k, v in _rse:
            rse_data[k] = v

    rse_data.pop('continent')
    rse_data.pop('ASN')
    rse_data.pop('ISP')
    rse_data.pop('deleted')
    rse_data.pop('deleted_at')

    # get RSE attributes
    rse_data['attributes'] = list_rse_attributes(rse_id=rse_id, session=session)

    protocols = get_rse_protocols(rse_id=rse_id, session=session)
    rse_data['lfn2pfn_algorithm'] = protocols.get('lfn2pfn_algorithm')
    rse_data['verify_checksum'] = protocols.get('verify_checksum')
    rse_data['credentials'] = protocols.get('credentials')
    rse_data['availability_delete'] = protocols.get('availability_delete')
    rse_data['availability_write'] = protocols.get('availability_write')
    rse_data['availability_read'] = protocols.get('availability_read')
    rse_data['protocols'] = protocols.get('protocols')

    # get RSE limits
    limits = get_rse_limits(rse_id=rse_id, session=session)
    rse_data['MinFreeSpace'] = limits.get('MinFreeSpace')

    return rse_data


@transactional_session
def add_qos_policy(
    rse_id: str,
    qos_policy: str,
    *,
    session: "Session"
) -> bool:
    """
    Add a QoS policy from an RSE.

    :param rse_id: The id of the RSE.
    :param qos_policy: The QoS policy to add.
    :param session: The database session in use.

    :raises Duplicate: If the QoS policy already exists.
    :returns: True if successful, except otherwise.
    """

    try:
        new_qos_policy = models.RSEQoSAssociation()
        new_qos_policy.update({'rse_id': rse_id,
                               'qos_policy': qos_policy})
        new_qos_policy.save(session=session)
    except (IntegrityError, FlushError, OperationalError) as error:
        if ('UNIQUE constraint failed' in error.args[0]) or ('conflicts with persistent instance' in error.args[0]) \
           or match('.*IntegrityError.*ORA-00001: unique constraint.*RSE_PROTOCOLS_PK.*violated.*', error.args[0]) \
           or match('.*IntegrityError.*1062.*Duplicate entry.*for key.*', error.args[0]) \
           or match('.*IntegrityError.*duplicate key value violates unique constraint.*', error.args[0])\
           or match('.*UniqueViolation.*duplicate key value violates unique constraint.*', error.args[0])\
           or match('.*IntegrityError.*columns.*are not unique.*', error.args[0]):
            raise exception.Duplicate('QoS policy %s already exists!' % qos_policy)
    except DatabaseError as error:
        raise exception.RucioException(error.args)

    return True


@transactional_session
def delete_qos_policy(
    rse_id: str,
    qos_policy: str,
    *,
    session: "Session"
) -> bool:
    """
    Delete a QoS policy from an RSE.

    :param rse_id: The id of the RSE.
    :param qos_policy: The QoS policy to delete.
    :param session: The database session in use.

    :returns: True if successful, silent failure if QoS policy does not exist.
    """

    try:
        stmt = delete(
            models.RSEQoSAssociation
        ).where(
            and_(models.RSEQoSAssociation.rse_id == rse_id,
                 models.RSEQoSAssociation.qos_policy == qos_policy)
        )
        session.execute(stmt)
    except DatabaseError as error:
        raise exception.RucioException(error.args)

    return True


@read_session
def list_qos_policies(
    rse_id: str,
    *,
    session: "Session"
) -> list[str]:
    """
    List all QoS policies of an RSE.

    :param rse_id: The id of the RSE.
    :param session: The database session in use.

    :returns: List containing all QoS policies.
    """

    qos_policies = []
    try:
        stmt = select(
            models.RSEQoSAssociation.qos_policy
        ).where(
            models.RSEQoSAssociation.rse_id == rse_id
        )
        for qos_policy in session.execute(stmt).scalars():
            qos_policies.append(qos_policy)
    except DatabaseError as error:
        raise exception.RucioException(error.args)

    return qos_policies


@transactional_session
def fill_rse_expired(
    rse_id: str,
    *,
    session: "Session"
) -> None:
    """
    Fill the rse_usage for source expired

    :param rse_id: The RSE id.
    """
    stmt = select(
        func.sum(models.RSEFileAssociation.bytes).label("bytes"),
        func.count().label("length")
    ).with_hint(
        models.RSEFileAssociation,
        'INDEX(REPLICAS REPLICAS_RSE_ID_TOMBSTONE_IDX)',
        'oracle'
    ).where(
        and_(models.RSEFileAssociation.tombstone < datetime.utcnow(),
             models.RSEFileAssociation.lock_cnt == 0,
             models.RSEFileAssociation.rse_id == rse_id,
             models.RSEFileAssociation.state.in_((ReplicaState.AVAILABLE, ReplicaState.UNAVAILABLE, ReplicaState.BAD)))
    )

    sum_bytes, sum_files = session.execute(stmt).one()
    models.RSEUsage(rse_id=rse_id,
                    used=sum_bytes,
                    files=sum_files,
                    source='expired').save(session=session)


def determine_audience_for_rse(rse_id: str) -> str:
    """Construct the Audience claim for an RSE."""
    rse_protocols = get_rse_protocols(rse_id)
    # FIXME: At the time of writing, there does not appear to be a common
    # agreement on how sites will configure their storages.  Rucio had requested
    # that the protocol hostname be sufficient, but this may not come to pass.
    filtered_hostnames = {p['hostname']
                          for p in rse_protocols['protocols']
                          if p['scheme'] == 'davs'}
    return ' '.join(sorted(filtered_hostnames))


def determine_scope_for_rse(
    rse_id: str,
    scopes: 'Iterable[str]',
    extra_scopes: Optional['Iterable[str]'] = None,
) -> str:
    """Construct the Scope claim for an RSE."""
    if extra_scopes is None:
        extra_scopes = []
    rse_protocols = get_rse_protocols(rse_id)
    filtered_prefixes = set()
    for protocol in rse_protocols['protocols']:
        # Token support is exclusive to WebDAV.
        if protocol['scheme'] != 'davs':
            continue
        # Remove base path from prefix.  Storages typically map an issuer (i.e.
        # a VO) to a particular area.  If so, then the path to that area acts as
        # a base which should be removed from the prefix (in order for '/' to
        # mean the entire resource associated with that issuer).
        prefix = protocol['prefix']
        if base_path := get_rse_attribute(rse_id, RseAttr.OIDC_BASE_PATH):  # type: ignore (session parameter missing)
            prefix = prefix.removeprefix(base_path)
        filtered_prefixes.add(prefix)
    all_scopes = [f'{s}:{p}' for s in scopes for p in filtered_prefixes] + list(extra_scopes)
    return ' '.join(sorted(all_scopes))



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rucio.common.utils import execute


class TestModuleImport:
    def test_import(self):
        """ """
        cmd = 'rucio --version'
        exitcode, out, err = execute(cmd)
        assert 'ImportError' not in err
        assert 'ImportError' not in out
        assert 'Exception' not in err
        assert 'Exception' not in out



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import datetime
import itertools
import logging
import threading
import weakref
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar, Union, cast

from sqlalchemy import and_, select

from rucio.common.config import config_get, config_get_int
from rucio.common.exception import InvalidRSEExpression, NoDistance, RSEProtocolNotSupported
from rucio.common.utils import PriorityQueue
from rucio.core.rse import RseCollection, RseData
from rucio.core.rse_expression_parser import parse_expression
from rucio.db.sqla import models
from rucio.db.sqla.session import read_session, transactional_session
from rucio.rse import rsemanager as rsemgr

_Number = Union[int, Decimal]
TN = TypeVar("TN", bound="Node")
TE = TypeVar("TE", bound="Edge")
ExpiringObjectCacheNewObject = TypeVar("ExpiringObjectCacheNewObject")

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator
    from typing import Protocol

    from sqlalchemy.orm import Session
    from typing_extensions import Self

    from rucio.common.types import HopDict, LoggerFunction

    class _StateProvider(Protocol):
        @property
        def cost(self) -> _Number:
            ...

        @property
        def enabled(self) -> bool:
            ...

    TNState = TypeVar("TNState", bound=_StateProvider)
    TEState = TypeVar("TEState", bound=_StateProvider)


DEFAULT_HOP_PENALTY = 10
INF = float('inf')


class Node(RseData):
    def __init__(self, rse_id: str) -> None:
        super().__init__(rse_id)

        self.in_edges = weakref.WeakKeyDictionary()
        self.out_edges = weakref.WeakKeyDictionary()

        self.cost: _Number = 0
        self.enabled: bool = True
        self.used_for_multihop = False


class Edge(Generic[TN]):
    def __init__(self, src_node: TN, dst_node: TN) -> None:
        self._src_node = weakref.ref(src_node)
        self._dst_node = weakref.ref(dst_node)

        self.cost: _Number = 1
        self.enabled: bool = True

        self.add_to_nodes()

    def add_to_nodes(self) -> None:
        self.src_node.out_edges[self.dst_node] = self
        self.dst_node.in_edges[self.src_node] = self

    def remove_from_nodes(self) -> None:
        self.src_node.out_edges.pop(self.dst_node, None)
        self.dst_node.in_edges.pop(self.src_node, None)

    @property
    def src_node(self) -> TN:
        node = self._src_node()
        if node is None:
            # This shouldn't happen if the Node list is correctly managed by the Topology object.
            raise ReferenceError("weak reference returned None")
        return node

    @property
    def dst_node(self) -> TN:
        node = self._dst_node()
        if node is None:
            # This shouldn't happen if the Node list is correctly managed by the Topology object.
            raise ReferenceError("weak reference returned None")
        return node

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._src_node == other._src_node and self._dst_node == other._dst_node

    def __str__(self) -> str:
        return f'{self._src_node}-->{self._dst_node}'


class Topology(RseCollection, Generic[TN, TE]):
    """
    Helper private class used to easily fetch topological information for a subset of RSEs.
    """
    def __init__(
            self,
            rse_ids: Optional["Iterable[str]"] = None,
            ignore_availability: bool = False,
            node_cls: type[TN] = Node,
            edge_cls: type[TE] = Edge,
    ) -> None:
        super().__init__(rse_ids=rse_ids, rse_data_cls=node_cls)
        self._edge_cls = edge_cls
        self._edges: dict[tuple[TN, TN], TE] = {}
        self._edges_loaded = False
        self._multihop_nodes = set()
        self._hop_penalty = DEFAULT_HOP_PENALTY
        self.ignore_availability = ignore_availability

        self._lock = threading.RLock()

    @transactional_session
    def ensure_loaded(
            self,
            rse_ids: "Optional[Iterable[str]]" = None,
            load_name: bool = False,
            load_columns: bool = False,
            load_attributes: bool = False,
            load_info: bool = False,
            load_usage: bool = False,
            load_limits: bool = False,
            include_deleted: bool = False,
            *,
            session: "Session",
    ) -> None:

        if not rse_ids:
            with self._lock:
                rse_ids = list(self.rse_id_to_data_map)
        super().ensure_loaded(
            rse_ids=rse_ids,
            load_name=load_name,
            load_columns=load_columns,
            load_attributes=load_attributes,
            load_info=load_info,
            load_usage=load_usage,
            load_limits=load_limits,
            include_deleted=include_deleted,
            session=session,
        )

    def get_or_create(self, rse_id: str) -> "TN":
        rse_data = self.rse_id_to_data_map.get(rse_id)
        if rse_data is None:
            with self._lock:
                rse_data = self.rse_id_to_data_map.get(rse_id)
                if not rse_data:
                    self.rse_id_to_data_map[rse_id] = rse_data = self._rse_data_cls(rse_id)
                    # A new node added. Edges which were already loaded are probably incomplete now.
                    self._edges_loaded = False
        return rse_data

    @property
    def edges(self) -> dict[tuple[TN, TN], TE]:
        with self._lock:
            return copy.copy(self._edges)

    def edge(self, src_node: TN, dst_node: TN) -> "Optional[TE]":
        return self._edges.get((src_node, dst_node))

    def get_or_create_edge(self, src_node: TN, dst_node: TN) -> "TE":
        edge = self._edges.get((src_node, dst_node))
        if not edge:
            with self._lock:
                edge = self._edges.get((src_node, dst_node))
                if not edge:
                    self._edges[src_node, dst_node] = edge = self._edge_cls(src_node, dst_node)
        return edge

    def delete_edge(self, src_node: TN, dst_node: TN) -> None:
        with self._lock:
            edge = self._edges[src_node, dst_node]
            edge.remove_from_nodes()

    @property
    def multihop_enabled(self) -> bool:
        return True if self._multihop_nodes else False

    @read_session
    def configure_multihop(self, multihop_rse_ids: Optional[set[str]] = None, *, session: "Session", logger: "LoggerFunction" = logging.log) -> "Self":
        with self._lock:
            return self._configure_multihop(multihop_rse_ids=multihop_rse_ids, session=session, logger=logger)

    def _configure_multihop(self, multihop_rse_ids: Optional[set[str]] = None, *, session: "Session", logger: "LoggerFunction" = logging.log) -> "Self":

        if multihop_rse_ids is None:
            multihop_rse_expression = config_get('transfers', 'multihop_rse_expression', default='available_for_multihop=true', expiration_time=600, session=session)

            multihop_rse_ids = set()
            if multihop_rse_expression.strip():
                try:
                    multihop_rse_ids = {rse['id'] for rse in parse_expression(multihop_rse_expression, session=session)}
                except InvalidRSEExpression:
                    pass
                if not multihop_rse_ids:
                    logger(logging.WARNING, 'multihop_rse_expression is not empty, but returned no RSEs')

        for node in self._multihop_nodes:
            node.used_for_multihop = False

        self._multihop_nodes.clear()

        for rse_id in multihop_rse_ids:
            node = self.get_or_create(rse_id).ensure_loaded(load_columns=True)
            if self.ignore_availability or (node.columns['availability_read'] and node.columns['availability_write']):
                node.used_for_multihop = True
                self._multihop_nodes.add(node)

        self._hop_penalty = config_get_int('transfers', 'hop_penalty', default=DEFAULT_HOP_PENALTY, session=session)
        return self

    @read_session
    def ensure_edges_loaded(self, *, session: "Session") -> None:
        """
        Ensure that all edges are loaded for the (sub-)set of nodes known by this topology object
        """
        if self._edges_loaded:
            return

        with self._lock:
            return self._ensure_edges_loaded(session=session)

    def _ensure_edges_loaded(self, *, session: "Session") -> None:
        stmt = select(
            models.Distance
        ).where(
            and_(
                models.Distance.src_rse_id.in_(self.rse_id_to_data_map.keys()),
                models.Distance.dest_rse_id.in_(self.rse_id_to_data_map.keys()),
            )
        )

        loaded_edges = set()
        for distance in session.execute(stmt).scalars():
            if distance.distance is None:
                continue

            src_node = self[distance.src_rse_id]
            dst_node = self[distance.dest_rse_id]
            edge = self.get_or_create_edge(src_node, dst_node)

            sanitized_dist = int(distance.distance) if distance.distance >= 0 else 0
            edge.cost = sanitized_dist

            loaded_edges.add((src_node, dst_node))

        if len(loaded_edges) != len(self._edges):
            # Remove edges which don't exist in the database anymore
            to_remove = set(self._edges).difference(loaded_edges)
            for src_node, dst_node in to_remove:
                self.delete_edge(src_node, dst_node)

        self._edges_loaded = True

    @read_session
    def search_shortest_paths(
            self,
            src_nodes: "Iterable[TN]",
            dst_node: TN,
            operation_src: str,
            operation_dest: str,
            domain: str,
            limit_dest_schemes: list[str],
            *,
            session: "Session",
    ) -> dict[TN, list[dict[str, Any]]]:
        """
        Find the shortest paths from multiple sources towards dest_rse_id.
        """

        for rse in itertools.chain(src_nodes, [dst_node], self._multihop_nodes):
            rse.ensure_loaded(load_attributes=True, load_info=True, session=session)
        self.ensure_edges_loaded(session=session)

        if self._multihop_nodes:
            # Filter out island source RSEs
            nodes_to_find = {node for node in src_nodes if node.out_edges}
        else:
            nodes_to_find = set(src_nodes)

        class _NodeStateProvider:
            _hop_penalty = self._hop_penalty

            def __init__(self, node: TN) -> None:
                self.enabled: bool = True
                self.cost: _Number = 0
                if node != dst_node:
                    try:
                        self.cost = int(node.attributes.get('hop_penalty', self._hop_penalty))
                    except ValueError:
                        self.cost = self._hop_penalty

        scheme_missmatch_found = {}

        class _EdgeStateProvider:
            def __init__(self, edge: TE) -> None:
                self.edge = edge
                self.chosen_scheme = {}

            @property
            def cost(self) -> _Number:
                return self.edge.cost

            @property
            def enabled(self) -> bool:
                try:
                    matching_scheme = rsemgr.find_matching_scheme(
                        rse_settings_src=self.edge.src_node.info,
                        rse_settings_dest=self.edge.dst_node.info,
                        operation_src=operation_src,
                        operation_dest=operation_dest,
                        domain=domain,
                        scheme=limit_dest_schemes if self.edge.dst_node == dst_node and limit_dest_schemes else None,
                    )
                    self.chosen_scheme = {
                        'source_scheme': matching_scheme[1],
                        'dest_scheme': matching_scheme[0],
                        'source_scheme_priority': matching_scheme[3],
                        'dest_scheme_priority': matching_scheme[2],
                    }
                    return True
                except RSEProtocolNotSupported:
                    scheme_missmatch_found[self.edge.src_node] = True
                    return False

        paths = {dst_node: []}
        for node, distance, _, edge_to_next_hop, edge_state in self.dijkstra_spf(dst_node=dst_node,
                                                                                 nodes_to_find=nodes_to_find,
                                                                                 node_state_provider=_NodeStateProvider,
                                                                                 edge_state_provider=_EdgeStateProvider):
            nh_node = edge_to_next_hop.dst_node
            edge_state = cast("_EdgeStateProvider", edge_state)
            hop = {
                'source_rse': node,
                'dest_rse': nh_node,
                'hop_distance': edge_state.cost,
                'cumulated_distance': distance,
                **edge_state.chosen_scheme,
            }
            paths[node] = [hop] + paths[nh_node]

            nodes_to_find.discard(node)
            if not nodes_to_find:
                # We found the shortest paths to all desired nodes
                break

        result = {}
        for node in src_nodes:
            path = paths.get(node)
            if path is not None:
                result[node] = path
            elif scheme_missmatch_found.get(node):
                result[node] = []
        return result

    def dijkstra_spf(
            self,
            dst_node: TN,
            nodes_to_find: Optional[set[TN]] = None,
            node_state_provider: "Callable[[TN], TNState]" = lambda x: x,
            edge_state_provider: "Callable[[TE], TEState]" = lambda x: x,
    ) -> "Iterator[tuple[TN, _Number, TNState, TE, TEState]]":
        """
        Does a Backwards Dijkstra's algorithm: start from destination and follow inbound links to other nodes.
        If multihop is disabled, stop after analysing direct connections to dest_rse.
        If the optional nodes_to_find parameter is set, will restrict search only towards these nodes.
        Otherwise, traverse the graph in integrality.

        Will yield nodes in order of their distance from the destination.
        """

        priority_q = PriorityQueue()
        priority_q[dst_node] = 0
        next_hops: dict[TN, tuple[_Number, TNState, Optional[TE], Optional[TEState]]] =\
            {dst_node: (0, node_state_provider(dst_node), None, None)}
        while priority_q:
            node = priority_q.pop()
            node_dist, node_state, edge_to_nh, edge_to_nh_state = next_hops[node]

            if edge_to_nh is not None and edge_to_nh_state is not None:  # skip dst_node
                yield node, node_dist, node_state, edge_to_nh, edge_to_nh_state

            if self._multihop_nodes or edge_to_nh is None:
                # If multihop is disabled, only examine neighbors of dst_node

                for adjacent_node, edge in node.in_edges.items():

                    if nodes_to_find is None or adjacent_node in nodes_to_find or adjacent_node.used_for_multihop:

                        edge_state = edge_state_provider(edge)
                        new_adjacent_dist = node_dist + node_state.cost + edge_state.cost
                        if new_adjacent_dist < next_hops.get(adjacent_node, (INF, ))[0] and edge_state.enabled:
                            adj_node_state = node_state_provider(adjacent_node)
                            next_hops[adjacent_node] = new_adjacent_dist, adj_node_state, edge, edge_state
                            priority_q[adjacent_node] = new_adjacent_dist


class ExpiringObjectCache(Generic[ExpiringObjectCacheNewObject]):
    """
    Thread-safe container which builds and object with the function passed in parameter and
    caches it for the TTL duration.
    """

    def __init__(
            self,
            ttl: int,
            new_obj_fnc: "Callable[[], ExpiringObjectCacheNewObject]"
    ) -> None:
        self._lock = threading.Lock()
        self._object: Optional[ExpiringObjectCacheNewObject] = None
        self._creation_time: Optional[datetime.datetime] = None
        self._new_obj_fnc = new_obj_fnc
        self._ttl = ttl

    def get(self, logger: "LoggerFunction" = logging.log) -> ExpiringObjectCacheNewObject:
        with self._lock:
            if not self._object \
                    or not self._creation_time \
                    or datetime.datetime.utcnow() - self._creation_time > datetime.timedelta(seconds=self._ttl):
                self._object = self._new_obj_fnc()
                self._creation_time = datetime.datetime.utcnow()
                logger(logging.INFO, "Refreshed topology object")
            return self._object


@transactional_session
def get_hops(
        source_rse_id: str,
        dest_rse_id: str,
        multihop_rse_ids: Optional[set[str]] = None,
        limit_dest_schemes: Optional[list[str]] = None,
        *, session: "Session",
) -> list["HopDict"]:
    """
    Get a list of hops needed to transfer date from source_rse_id to dest_rse_id.
    Ideally, the list will only include one item (dest_rse_id) since no hops are needed.
    :param source_rse_id:       Source RSE id of the transfer.
    :param dest_rse_id:         Dest RSE id of the transfer.
    :param multihop_rse_ids:    List of RSE ids that can be used for multihop. If empty, multihop is disabled.
    :param limit_dest_schemes:  List of destination schemes the matching scheme algorithm should be limited to for a single hop.
    :returns:                   List of hops in the format [{'source_rse_id': source_rse_id, 'source_scheme': 'srm', 'source_scheme_priority': N, 'dest_rse_id': dest_rse_id, 'dest_scheme': 'srm', 'dest_scheme_priority': N}]
    :raises:                    NoDistance
    """
    if not limit_dest_schemes:
        limit_dest_schemes = []

    topology = Topology().configure_multihop(multihop_rse_ids=multihop_rse_ids)
    src_node = topology[source_rse_id]
    dst_node = topology[dest_rse_id]
    shortest_paths = topology.search_shortest_paths(src_nodes=[src_node], dst_node=dst_node,
                                                    operation_src='third_party_copy_read', operation_dest='third_party_copy_write',
                                                    domain='wan', limit_dest_schemes=limit_dest_schemes, session=session)

    path = shortest_paths.get(src_node)
    if path is None:
        raise NoDistance()

    if not path:
        raise RSEProtocolNotSupported()

    return path



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from unittest.mock import MagicMock

import pytest

from rucio.core import distance as distance_core
from rucio.core import replica as replica_core
from rucio.core import rule as rule_core
from rucio.core.request import list_and_mark_transfer_requests_and_source_replicas
from rucio.core.topology import Topology
from rucio.core.transfer import ProtocolFactory, build_transfer_paths
from rucio.db.sqla.session import get_session
from rucio.transfertool.fts3 import FTS3Transfertool, build_job_params
from rucio.transfertool.fts3_plugins import FTS3TapeMetadataPlugin

mock_session = get_session()

MAX_POLL_WAIT_SECONDS = 60
TEST_FTS_HOST = "https://fts:8446"


def _make_transfer_path(
        did,
        rse_factory,
        root_account,
        src_rse_is_tape: bool = False,
        dst_rse_is_tape: bool = True
):
    _, src_rse_id = rse_factory.make_mock_rse(rse_type="TAPE" if src_rse_is_tape else "DISK")
    dst_rse, dst_rse_id = rse_factory.make_mock_rse(rse_type="TAPE" if dst_rse_is_tape else "DISK")
    all_rses = [src_rse_id, dst_rse_id]
    distance_core.add_distance(src_rse_id, dst_rse_id, distance=1)

    topology = Topology(rse_ids=all_rses)
    file = {"scope": did["scope"], "name": did["name"], "bytes": 1}
    replica_core.add_replicas(rse_id=src_rse_id, files=[file], account=root_account)
    rule_core.add_rule(
        dids=[did],
        account=root_account,
        copies=1,
        rse_expression=dst_rse,
        grouping="ALL",
        weight=None,
        lifetime=None,
        locked=False,
        subscription_id=None,
    )

    requests_by_id = list_and_mark_transfer_requests_and_source_replicas(
        rse_collection=topology, rses=[src_rse_id, dst_rse_id]
    )
    requests, *_ = build_transfer_paths(
        topology=topology,
        protocol_factory=ProtocolFactory(),
        requests_with_sources=requests_by_id.values(),
    )
    _, [transfer_path] = next(iter(requests.items()))

    return transfer_path


@pytest.mark.parametrize("file_config_mock", [
    {
        "overrides": [
            ('tape_priority', 'User Subscriptions', '100'),
            ("transfers", "fts3tape_metadata_plugins", "activity")
        ]
    }
], indirect=True)
def test_scheduling_hints(file_config_mock, did_factory, rse_factory, root_account):

    """Transfer with an installed plugin for scheduling based on transfer activity"""

    # Produce a new did
    mock_did = did_factory.random_file_did()

    # Make the transfer path
    transfer_path = _make_transfer_path(mock_did, rse_factory, root_account)

    # Mock Transfer Tool
    fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

    job_params = build_job_params(
        transfer_path=transfer_path,
        bring_online=None,
        default_lifetime=None,
        archive_timeout_override=None,
        max_time_in_queue=None,
        logger=logging.log,
    )
    # Get the job params used for each transfer
    job_params = fts3_tool._file_from_transfer(transfer_path[0], job_params)

    # Extract the hints
    assert "archive_metadata" in job_params
    assert len(job_params["archive_metadata"].keys()) == 1
    generated_scheduling_hints = job_params["archive_metadata"]["scheduling_hints"]

    expected_scheduling_hints = {"priority": 100}
    assert expected_scheduling_hints == generated_scheduling_hints


@pytest.mark.parametrize("file_config_mock", [
    {
        "overrides": [
            ("transfers", "fts3tape_metadata_plugins", "activity")
        ]
    }
], indirect=True)
def test_activity_missing(file_config_mock, did_factory, rse_factory, root_account):
    """Ensure default is selected when the activity is not listed in the config, but is in the schema"""
    # Do not add config section for priority, but do add the FTS3
    # Produce a new did
    mock_did = did_factory.random_file_did()

    # Make the transfer path
    transfer_path = _make_transfer_path(mock_did, rse_factory, root_account)

    # Mock Transfer Tool
    fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

    job_params = build_job_params(
        transfer_path=transfer_path,
        bring_online=None,
        default_lifetime=None,
        archive_timeout_override=None,
        max_time_in_queue=None,
        logger=logging.log,
    )

    # Get the job params used for each transfer
    transfer_params = transfer_path[0]
    transfer_params.rws.activity = "Not A Real Activity"

    job_params = fts3_tool._file_from_transfer(transfer_params, job_params)

    # Extract the hints
    assert "archive_metadata" in job_params
    generated_scheduling_hints = job_params["archive_metadata"]["scheduling_hints"]

    expected_scheduling_hints = {"priority": 20}
    assert expected_scheduling_hints == generated_scheduling_hints


class TestCollocationHints:
    class TestCollocationPlugin(FTS3TapeMetadataPlugin):
        def __init__(self) -> None:
            self.register(
                'test',
                func=lambda x: self._test_collocation(x))
            super().__init__('test')

        def _test_collocation(self, hints: dict[str, str]) -> dict[str, dict]:
            return {"collocation_hints": {"0": "", "1": "", "2": "", "3": ""}}

    TestCollocationPlugin()

    @pytest.mark.parametrize("file_config_mock", [
        {
            "overrides": [
                ("transfers", "fts3tape_metadata_plugins", "test")
            ]
        }
    ], indirect=True)
    def test_collocation_hints(self, file_config_mock, did_factory, rse_factory, root_account):
        """For a mock collocation algorithm, it can produce the 4 levels of hints required for each did"""

        mock_did = did_factory.random_file_did()
        transfer_path = _make_transfer_path(mock_did, rse_factory, root_account)

        # Mock Transfer Tool
        fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

        job_params = build_job_params(
            transfer_path=transfer_path,
            bring_online=None,
            default_lifetime=None,
            archive_timeout_override=None,
            max_time_in_queue=None,
            logger=logging.log,
        )

        # Get the job params used for each transfer
        job_params = fts3_tool._file_from_transfer(transfer_path[0], job_params)

        expected_collocation_hints = {
            "collocation_hints": {
                "0": "",
                "1": "",
                "2": "",
                "3": "",
            }
        }

        assert "archive_metadata" in job_params
        generated_collocation_hints = job_params["archive_metadata"]["collocation_hints"]

        assert (
            expected_collocation_hints["collocation_hints"] == generated_collocation_hints
        )

    @pytest.mark.parametrize("file_config_mock", [
        {
            "overrides": [
                ("transfers", "fts3tape_metadata_plugins", "activity, test")
            ]
        }
    ], indirect=True)
    def test_multiple_plugin_concat(self, file_config_mock, did_factory, rse_factory, root_account):
        """When multiple plugins are used (like priority and collocation), both logics are applied"""

        mock_did = did_factory.random_file_did()
        transfer_path = _make_transfer_path(mock_did, rse_factory, root_account)

        # Mock Transfer Tool
        fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

        job_params = build_job_params(
            transfer_path=transfer_path,
            bring_online=None,
            default_lifetime=None,
            archive_timeout_override=None,
            max_time_in_queue=None,
            logger=logging.log,
        )

        # Get the job params used for each transfer
        job_params = fts3_tool._file_from_transfer(transfer_path[0], job_params)
        expected_hints = {
            "scheduling_hints": {"priority": 20},
            "collocation_hints": {"0": "", "1": "", "2": "", "3": ""},
        }
        assert "archive_metadata" in job_params

        generated_collocation_hints = job_params["archive_metadata"]["collocation_hints"]
        assert expected_hints["collocation_hints"] == generated_collocation_hints

        generated_scheduling_hints = job_params["archive_metadata"]["scheduling_hints"]
        assert expected_hints["scheduling_hints"] == generated_scheduling_hints


@pytest.mark.parametrize("file_config_mock", [
    {
        "overrides": [
            ("transfers", "metadata_byte_limit", "4"),
            ("transfers", "fts3tape_metadata_plugins", "def")

        ]
    }
], indirect=True)
def test_transfer_over_limit(file_config_mock, did_factory, rse_factory, root_account):
    mock_did = did_factory.random_file_did()
    transfer_path = _make_transfer_path(mock_did, rse_factory, root_account)

    # Mock Transfer Tool
    fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

    job_params = build_job_params(
        transfer_path=transfer_path,
        bring_online=None,
        default_lifetime=None,
        archive_timeout_override=None,
        max_time_in_queue=None,
        logger=logging.log,
    )

    from rucio.common.exception import InvalidRequest

    with pytest.raises(InvalidRequest):
        job_params = fts3_tool._file_from_transfer(transfer_path[0], job_params)


def test_include_none_by_default(did_factory, rse_factory, root_account):
    # When there is not plugin specified in the config, do not include any archive metadata

    mock_did = did_factory.random_file_did()
    transfer_path = _make_transfer_path(mock_did, rse_factory, root_account)

    # Mock Transfer Tool
    fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

    job_params = build_job_params(
        transfer_path=transfer_path,
        bring_online=None,
        default_lifetime=None,
        archive_timeout_override=None,
        max_time_in_queue=None,
        logger=logging.log,
    )

    job_params = fts3_tool._file_from_transfer(transfer_path[0], job_params)
    assert "archive_metadata" not in job_params


def test_no_metadata_if_dest_is_not_tape(did_factory, rse_factory, root_account):
    """
    If the destination RSE is not tape,
    no tape metadata plugin should run,
    and no archive metadata should be included.
    """

    mock_did = did_factory.random_file_did()
    mock_metadata_plugin = MagicMock()
    transfer_path = _make_transfer_path(mock_did, rse_factory, root_account, dst_rse_is_tape=False)

    fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

    fts3_tool.tape_metadata_plugins = [mock_metadata_plugin]  # type: ignore

    job_params = build_job_params(
        transfer_path=transfer_path,
        bring_online=None,
        default_lifetime=None,
        archive_timeout_override=None,
        max_time_in_queue=None,
        logger=logging.log,
    )

    job_params = fts3_tool._file_from_transfer(transfer_path[0], job_params)

    assert not mock_metadata_plugin.called
    assert "archive_metadata" not in job_params


def test_staging_metadata_if_src_is_tape(did_factory, rse_factory, root_account):
    """
    If the source RSE is tape,
    staging metadata should be included.
    """

    mock_did = did_factory.random_file_did()
    transfer_path = _make_transfer_path(mock_did, rse_factory, root_account, src_rse_is_tape=True)

    fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

    job_params = build_job_params(
        transfer_path=transfer_path,
        bring_online=None,
        default_lifetime=None,
        archive_timeout_override=None,
        max_time_in_queue=None,
        logger=logging.log,
    )

    # Get the job params used for each transfer
    transfer_params = transfer_path[0]
    transfer_params.rws.activity = "Not A Real Activity"

    job_params = fts3_tool._file_from_transfer(transfer_params, job_params)

    actual_staging_metadata = job_params.get("staging_metadata", None)
    assert actual_staging_metadata == {"default": {"activity": "Not A Real Activity"}}


def test_no_staging_metadata_if_src_is_not_tape(did_factory, rse_factory, root_account):
    """
    If the source RSE is not tape,
    staging metadata should not be included.
    """

    mock_did = did_factory.random_file_did()
    transfer_path = _make_transfer_path(mock_did, rse_factory, root_account, src_rse_is_tape=False)

    fts3_tool = FTS3Transfertool(TEST_FTS_HOST)

    job_params = build_job_params(
        transfer_path=transfer_path,
        bring_online=None,
        default_lifetime=None,
        archive_timeout_override=None,
        max_time_in_queue=None,
        logger=logging.log,
    )

    # Get the job params used for each transfer
    transfer_params = transfer_path[0]
    transfer_params.rws.activity = "Not A Real Activity"

    job_params = fts3_tool._file_from_transfer(transfer_params, job_params)

    assert "staging_metadata" not in job_params



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


'''
 Elasticsearch based metadata plugin.
'''

import datetime
import operator
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from elasticsearch import Elasticsearch
from elasticsearch import exceptions as elastic_exceptions

from rucio.common import config, exception
from rucio.core.did_meta_plugins.did_meta_plugin_interface import DidMetaPlugin
from rucio.core.did_meta_plugins.filter_engine import FilterEngine

if TYPE_CHECKING:
    from collections.abc import Iterator

    from sqlalchemy.orm import Session

    from rucio.common.types import InternalScope

IMMUTABLE_KEYS = [
    'scope',            # generated on insert
    'name',             # generated on insert
    'vo'                # generated on insert
]


class ElasticDidMeta(DidMetaPlugin):
    def __init__(
        self,
        hosts: Optional[list[str]] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        index: Optional[str] = None,
        archive_index: Optional[str] = None,
        use_ssl: Optional[bool] = False,
        verify_certs: bool = True,
        ca_certs: Optional[str] = None,
        client_cert: Optional[str] = None,
        client_key: Optional[str] = None,
        request_timeout: int = 100,
        max_retries: int = 3,
        retry_on_timeout: bool = False
    ) -> None:
        super(ElasticDidMeta, self).__init__()
        hosts = hosts or [config.config_get('metadata', 'elastic_service_hosts')]
        user = user or config.config_get('metadata', 'elastic_user', False, None)
        password = password or config.config_get('metadata', 'elastic_password', False, None)
        self.index = index or config.config_get('metadata', 'meta_index', False, 'rucio_did_meta')
        self.archive_index = archive_index or config.config_get('metadata', 'archive_index', False, 'archive_meta')
        use_ssl = use_ssl or config.config_get_bool('metadata', 'use_ssl', False, False)
        ca_certs = ca_certs or config.config_get('metadata', 'ca_certs', False, None)
        client_cert = client_cert or config.config_get('metadata', 'client_cert', False, None)
        client_key = client_key or config.config_get('metadata', 'client_key', False, None)

        self.es_config = {
            'hosts': hosts,
            'timeout': request_timeout,
            'max_retries': max_retries,
            'retry_on_timeout': retry_on_timeout
        }
        if user and password:
            self.es_config['basic_auth'] = (user, password)

        if use_ssl:
            self.es_config.update({
                'ca_certs': ca_certs,
                'verify_certs': verify_certs,
            })
            if client_cert and client_key:
                self.es_config.update({
                    'client_cert': client_cert,
                    'client_key': client_key
                })

        self.client = Elasticsearch(**self.es_config)
        self.plugin_name = "ELASTIC"

    def drop_index(self) -> None:
        self.client.indices.delete(index=self.index)

    def get_metadata(
        self,
        scope: "InternalScope",
        name: str,
        *,
        session: "Optional[Session]" = None
    ) -> dict[str, Any]:
        """
        Get data identifier metadata.

        :param scope: The scope name
        :param name: The data identifier name
        :param session: The database session in use
        :returns: The metadata for the did
        :raises DataIdentifierNotFound: If the DID metadata is not found.
        :raises RucioException: If another error occurs during the process.
        """

        doc_id = f"{scope.internal}{name}"
        try:
            doc = self.client.get(index=self.index, id=doc_id)["_source"]
        except elastic_exceptions.NotFoundError as err:
            raise exception.DataIdentifierNotFound(f"No metadata found for DID '{scope}:{name}' not found") from err
        except Exception as err:
            raise exception.RucioException(err)
        return doc

    def set_metadata(
        self,
        scope: "InternalScope",
        name: str,
        key: str,
        value: str,
        recursive: bool = False,
        *,
        session: "Optional[Session]" = None
    ) -> None:
        """
        Set single metadata key.

        :param scope: the scope of did
        :param name: the name of the did
        :param key: the key to be added
        :param value: the value of the key to be added
        :param recursive: recurse into DIDs (not supported)
        :param session: The database session in use
        :raises DataIdentifierNotFound: If the DID is not found.
        :raises RucioException: If an error occurs while setting the metadata.
        """
        self.set_metadata_bulk(scope=scope, name=name, meta={key: value}, recursive=recursive, session=session)

    def set_metadata_bulk(
        self,
        scope: "InternalScope",
        name: str,
        meta: dict[str, Any],
        recursive: bool = False,
        *,
        session: "Optional[Session]" = None
    ) -> None:
        """
        Bulk set metadata keys.

        :param scope: the scope of did
        :param name: the name of the did
        :param meta: dictionary of metadata keypairs to be added
        :param recursive: recurse into DIDs (not supported)
        :param session: The database session in use
        :raises DataIdentifierNotFound: If the DID is not found.
        :raises UnsupportedOperation: If recursive inserts are requested (currently unsupported).
        :raises RucioException: If an error occurs while setting the metadata.
        """
        doc_id = f"{scope.internal}{name}"
        try:
            # Try to get existing metadata
            existing_meta = self.get_metadata(scope, name)
        except exception.DataIdentifierNotFound:
            existing_meta = {
                'scope': str(scope.external),
                'name': name,
                'vo': scope.vo
            }
        for key, value in meta.items():
            if key not in IMMUTABLE_KEYS:
                existing_meta[key] = value

        try:
            self.client.index(index=self.index, body=existing_meta, id=doc_id, refresh="true")
        except Exception as err:
            raise exception.RucioException(err)

        if recursive:
            raise exception.UnsupportedOperation(f"'{self.plugin_name.lower()}' metadata module does not currently support recursive inserts of metadata")

    def delete_metadata(
        self,
        scope: "InternalScope",
        name: str,
        key: str,
        *,
        session: "Optional[Session]" = None
    ) -> None:
        """
        Delete a key from metadata.

        :param scope: the scope of did
        :param name: the name of the did
        :param key: the key to be deleted
        :raises DataIdentifierNotFound: If the DID is not found.
        :raises RucioException: If an error occurs while setting the metadata.
        """
        doc_id = f"{scope.internal}{name}"
        try:
            # First, get the current document
            doc = self.client.get(index=self.index, id=doc_id)

            # Check if the key exists in the document
            if key in doc['_source']:
                # Use script to remove the field
                script = {
                    "script": {
                        "source": f"ctx._source.remove('{key}')",
                        "lang": "painless"
                    }
                }
                self.client.update(index=self.index, id=doc_id, body=script)
        except elastic_exceptions.NotFoundError as err:
            raise exception.DataIdentifierNotFound(f"No metadata found for DID '{scope}:{name}' not found") from err
        except Exception as err:
            raise exception.RucioException(err)

    def list_dids(
        self,
        scope: "InternalScope",
        filters: Union[list[dict[str, Any]], dict[str, Any]],
        did_type: Literal['all', 'collection', 'dataset', 'container', 'file'] = 'collection',
        ignore_case: bool = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        long: bool = False,
        recursive: bool = False,
        ignore_dids: Optional[list] = None,
        *,
        session: "Optional[Session]" = None
    ) -> "Iterator[dict[str, Any]]":
        """
        List DIDs (Data Identifier).

        :param scope: The scope of the DIDs to search.
        :param filters: The filters to apply to the DID search.
        :param did_type: The type of DID (default is 'collection').
        :param ignore_case: Whether to ignore case (default is False).
        :param limit: The maximum number of DIDs to return.
        :param offset: The starting point for the search (used for pagination).
        :param long: Whether to return extended information (scope, name, did_type, bytes, length) (default is False).
        :param recursive: Whether to search recursively (currently unsupported).
        :param ignore_dids: A list of DIDs to ignore (default is an empty list).
        :param session: The database session in use.
        :returns: A generator yielding DIDs as strings (when `long` is False) or dictionaries (when `long` is True).
        :raises UnsupportedOperation: If recursive searches are requested (currently unsupported).
        :raises RucioException: If an error occurs during the search.
        """

        if not ignore_dids:
            ignore_dids = []

        # backwards compatibility for filters as single {}.
        if isinstance(filters, dict):
            filters = [filters]

        # Create Elasticsearch query
        fe = FilterEngine(filters, model_class=None, strict_coerce=False)
        elastic_query_str = fe.create_elastic_query(
            additional_filters=[
                ('scope', operator.eq, str(scope.external)),
                ('vo', operator.eq, scope.vo)
            ]
        )
        pit = self.client.open_point_in_time(index=self.index, keep_alive="2m")
        pit_id = pit["id"]
        # Base query with point in time (pit) paramter.
        # sort is needed for search_after, so we use scope sort (random choice)
        query = {
            "query": elastic_query_str,
            "sort": [{"scope.keyword": "asc"}],
            "_source": ["scope", "name"] if not long else ["scope", "name", "did_type", "bytes", "length"],
            "pit": {"id": pit_id, "keep_alive": "2m"}
        }

        # Add sorting and pagination
        if offset:
            query["from"] = offset
        size = limit if limit else 10000
        query["size"] = size
        search_after = None
        total_processed = 0
        try:
            while True:
                if search_after:
                    query["search_after"] = search_after
                    query.pop("from", None)
                # Execute search
                results = self.client.search(body=query)
                hits = results['hits']['hits']
                if not hits:
                    break

                for hit in hits:
                    did_full = f"{hit['_source']['scope']}:{hit['_source']['name']}"
                    if did_full not in ignore_dids:
                        ignore_dids.append(did_full)
                        if long:
                            yield {
                                'scope': (hit['_source']['scope']),
                                'name': hit['_source']['name'],
                                'did_type': hit['_source'].get('did_type', 'N/A'),
                                'bytes': hit['_source'].get('bytes', 'N/A'),
                                'length': hit['_source'].get('length', 'N/A')
                            }
                        else:
                            yield hit['_source']['name']

                    total_processed += 1
                    if limit and total_processed >= limit:
                        break

                # Update search_after for the next iteration
                search_after = hits[-1]["sort"]

        finally:
            # Always delete the point in time when done
            self.client.close_point_in_time(body={"id": pit_id})

        if recursive:
            raise exception.UnsupportedOperation(f"'{self.plugin_name.lower()}' metadata module does not currently support recursive searches")

    def on_delete(
        self,
        scope: "InternalScope",
        name: str,
        archive: bool = False,
        session: "Optional[Session]" = None
    ) -> None:
        """
        Delete a document and optionally archive it.

        :param scope: The scope of the document
        :param name: The name of the document
        :param archive: Whether to archive the document before deletion
        :raises DataIdentifierNotFound: If the DID is not found.
        :raises RucioException: If an error occurs while setting the metadata.
        """
        doc_id = f"{scope}{name}"

        try:
            doc = self.client.get(index=self.index, id=doc_id)

            if archive:
                archived_doc = doc['_source']
                archived_doc['deleted_at'] = datetime.datetime.now(datetime.timezone.utc).isoformat()
                self.client.index(index=self.archive_index, id=doc_id, body=archived_doc)

            self.client.delete(index=self.index, id=doc_id)

        except elastic_exceptions.NotFoundError as err:
            raise exception.DataIdentifierNotFound(f"No metadata found for DID '{scope}:{name}' not found") from err
        except Exception as err:
            raise exception.RucioException(err)

    def get_metadata_archived(
        self,
        scope: "InternalScope",
        name: str,
        session: "Optional[Session]" = None
    ) -> None:
        """
        Retrieve archived metadata for a given scope and name.

        :param scope: The scope of the document
        :param name: The name of the document
        :return: The archived metadata or None if not found
        :raises DataIdentifierNotFound: If the DID is not found.
        :raises RucioException: If an error occurs while setting the metadata.
        """
        doc_id = f"{scope}{name}"

        try:
            doc = self.client.get(index=self.archive_index, id=doc_id)["_source"]
            return doc
        except elastic_exceptions.NotFoundError as err:
            raise exception.DataIdentifierNotFound(f"No metadata found for DID '{scope}:{name}' not found") from err
        except Exception as err:
            raise exception.RucioException(err)

    def manages_key(
        self,
        key: str,
        *,
        session: "Optional[Session]" = None
    ) -> bool:
        return True

    def get_plugin_name(self) -> str:
        """
        Returns a unique identifier for this plugin. This can be later used for filtering down results to this plugin only.

        :returns: The name of the plugin
        """
        return self.plugin_name



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import click

from rucio.cli.bin_legacy.rucio import download as download_exe
from rucio.cli.utils import Arguments
from rucio.common.config import config_get_float


@click.command()
@click.argument("dids", nargs=-1)
@click.option("--dir", default=".", help="The directory to store the downloaded file.")
@click.option("--allow-tape", is_flag=True, default=False, help="Also consider tape endpoints as source of the download.")
@click.option("--rses", "--rse-exp", help="RSE Expression to specify allowed sources")
@click.option("--impl", help="Transfer protocol implementation to use (e.g: xrootd, gfal.NoRename, webdav, ssh.Rsync, rclone).")
@click.option("--protocol", help="Force the protocol to use.")
@click.option("--nrandom", type=int, help="Download N random files from the DID.")
@click.option("--ndownloader", type=int, default=3, help="Choose the number of parallel processes for download.")
@click.option("--no-subdir", is_flag=True, default=False, help="Don't create a subdirectory for the scope of the files.")
@click.option("--pfn", help="Specify the exact PFN for the download.")
@click.option("--no-resolve-archives", is_flag=True, default=False, help="If set archives will not be considered for download.")
@click.option("--ignore-checksum", is_flag=True, default=False, help="Don't validate checksum for downloaded files.")
@click.option("--check-local-with-filesize-only", is_flag=True, default=False, help="Don't use checksum verification for already downloaded files, use filesize instead.")
@click.option(
    "--transfer-timeout",
    type=float,
    default=config_get_float("download", "transfer_timeout", False, None),
    help="Transfer timeout (in seconds). Default: computed dynamically from --transfer-speed-timeout. If set to any value >= 0, --transfer-speed-timeout is ignored.",
)  # NOQA: E501
@click.option("--transfer-speed-timeout", type=float, default=None, help="Minimum allowed average transfer speed (in KBps). Default: 500. Used to dynamically compute the timeout if --transfer-timeout not set. Is not supported for --pfn.")  # NOQA: E501
@click.option("--aria", default=False, help="Use aria2c utility if possible. (EXPERIMENTAL)")
@click.option("--archive-did", hidden=True)  # Download from archive is transparent. This option is obsolete.
# TODO replace with click functionality to automatically pull env vars
@click.option("--trace-appid", default=os.environ.get("RUCIO_TRACE_APPID", None), hidden=True)
@click.option("--trace-dataset", default=os.environ.get("RUCIO_TRACE_DATASET", None), hidden=True)
@click.option("--trace-datasetscope", default=os.environ.get("RUCIO_TRACE_DATASETSCOPE", None), hidden=True)
@click.option("--trace-eventtype", default=os.environ.get("RUCIO_TRACE_EVENTTYPE", None), hidden=True)
@click.option("--trace-pq", default=os.environ.get("RUCIO_TRACE_PQ", None), hidden=True)
@click.option("--trace-taskid", default=os.environ.get("RUCIO_TRACE_TASKID", None), hidden=True)
@click.option("--trace-usrdn", default=os.environ.get("RUCIO_TRACE_USRDN", None), hidden=True)
@click.option("--filter", help="Filter files by key-value pairs like guid=2e2232aafac8324db452070304f8d745.")
@click.option("--scope", help="Scope if you are using the filter option and no full DID.")
@click.option("--metalink", help="Path to a metalink file.")
@click.option("--no-show-download-exceptions", default=False, is_flag=True, help="Does not raise NoFilesDownloaded, NotAllFilesDownloaded or incorrect number of output queue files Exception.")  # NOQA: E501
@click.option("--replica-selection", help="Select the best replica using a replica sorting algorithm provided by replica sorter (e.g., random, geoip).")
@click.pass_context
def download(
    ctx,
    dids,
    dir,
    allow_tape,
    rses,
    impl,
    protocol,
    nrandom,
    ndownloader,
    no_subdir,
    pfn,
    no_resolve_archives,
    ignore_checksum,
    check_local_with_filesize_only,
    transfer_timeout,
    transfer_speed_timeout,
    aria,
    archive_did,
    trace_appid,
    trace_dataset,
    trace_datasetscope,
    trace_eventtype,
    trace_pq,
    trace_taskid,
    trace_usrdn,
    filter,
    scope,
    metalink,
    no_show_download_exceptions,
    replica_selection,
):
    """
    Download DID(s) (in the form of scope:name) to a local dir
    """
    args = Arguments(
        {
            "dids": dids,
            "dir": dir,
            "allow_tape": allow_tape,
            "rses": rses,
            "impl": impl,
            "protocol": protocol,
            "nrandom": nrandom,
            "ndownloader": ndownloader,
            "no_subdir": no_subdir,
            "pfn": pfn,
            "no_resolve_archives": no_resolve_archives,
            "ignore_checksum": ignore_checksum,
            "check_local_with_filesize_only": check_local_with_filesize_only,
            "transfer_timeout": transfer_timeout,
            "transfer_speed_timeout": transfer_speed_timeout,
            "aria": aria,
            "archive_did": archive_did,
            "trace_appid": trace_appid,
            "trace_dataset": trace_dataset,
            "trace_datasetscope": trace_datasetscope,
            "trace_eventtype": trace_eventtype,
            "trace_pq": trace_pq,
            "trace_taskid": trace_taskid,
            "trace_usrdn": trace_usrdn,
            "filter": filter,
            "scope": scope,
            "metalink_file": metalink,
            "deactivate_file_download_exceptions": no_show_download_exceptions,
            "sort": replica_selection,
        }
    )
    download_exe(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class TestBoolean:

    def test_booleanstring_account_attribute(self, random_account, account_client):
        account = random_account.external
        account_client.add_account_attribute(account, 'teststringtrue', 'true')
        account_client.add_account_attribute(account, 'testinttrue', '1')

        account_client.add_account_attribute(account, 'teststringfalse', 'false')
        account_client.add_account_attribute(account, 'testintfalse', '0')

        result = {}
        for account in account_client.list_account_attributes(account):
            for res in account:
                result[res['key']] = res['value']

        assert result['teststringtrue'] is True
        assert result['testinttrue'] == '1'
        assert result['teststringfalse'] is False
        assert result['testintfalse'] == '0'

    def test_booleanstring_rse_attribute(self, rse_client, rse_factory):
        rse, _ = rse_factory.make_mock_rse()
        rse_client.add_rse_attribute(rse, 'teststringtrue', 'true')
        rse_client.add_rse_attribute(rse, 'testinttrue', '1')

        rse_client.add_rse_attribute(rse, 'teststringfalse', 'false')
        rse_client.add_rse_attribute(rse, 'testintfalse', '0')

        result = rse_client.list_rse_attributes(rse)

        assert result['teststringtrue'] is True
        assert result['testinttrue'] == '1'
        assert result['teststringfalse'] is False
        assert result['testintfalse'] == '0'



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import tempfile

import pytest

from rucio.common.exception import RucioException
from rucio.common.utils import generate_uuid
from rucio.tests.common import account_name_generator, execute, file_generator, rse_name_generator, scope_name_generator


def test_main_args():
    specify_account = "rucio --account root --auth-strategy userpass whoami"
    exitcode, out, err = execute(specify_account)
    assert exitcode == 0
    assert "This method is being deprecated" not in err
    assert "root" in out

    legacy_arg = "rucio --legacy --account root --auth-strategy userpass whoami"
    exitcode, out, err = execute(legacy_arg)

    assert exitcode == 0
    assert "This method is being deprecated" in err
    assert "root" in out


def test_help_menus():
    """Verify help menus"""
    exitcode, out, err = execute("rucio --help")
    assert exitcode == 0
    assert "ERROR" not in err
    out = out.split("\n")
    commands = [cmd.split(" ") for cmd in out[out.index("Commands:")+1:] if len(cmd) > 3]  # command has two spaces in front of it
    commands = [cmd[2] for cmd in commands]

    for command in commands:
        exitcode, out, err = execute(f"rucio {command} --help")
        assert exitcode == 0, f"Command {command} --help failed"  # Included for debugging purposes

        exitcode, out, err = execute(f"rucio {command} -h")
        assert exitcode == 0, f"Command {command} -h failed"

        # test the subcommands/operations as well
        out = out.split("\n")
        try:
            subcommands = [cmd.split(" ") for cmd in out[out.index("Commands:")+1:] if len(cmd) > 3]
            subcommands = [cmd[2] for cmd in subcommands]
            for subcommand in subcommands:
                menu = f"rucio {command} {subcommand} --help"
                exitcode, out, err = execute(menu)
                assert exitcode == 0, f"Command {menu} failed"  # Included for debugging purposes

                menu = f"rucio {command} {subcommand} -h"
                exitcode, out, err = execute(menu)
                assert exitcode == 0, f"Command {menu} -h failed"

        except ValueError:  # "Commands:" not in list
            continue


def test_account(rucio_client):
    new_account = account_name_generator()
    command = f"rucio account add {new_account} USER"
    exitcode, _, err = execute(command)
    assert exitcode == 0
    assert "ERROR" not in err
    assert new_account in [account["account"] for account in rucio_client.list_accounts()]

    update_email = "jdoe@cern.ch"
    command = f"rucio account update {new_account} --email {update_email}"
    exitcode, _, err = execute(command)
    assert exitcode == 0
    assert "ERROR" not in err
    assert update_email == [account["email"] for account in rucio_client.list_accounts() if account["account"] == new_account][0]

    command = "rucio account list"
    exitcode, out, err = execute(command)
    assert exitcode == 0
    assert "ERROR" not in err
    assert new_account in out

    command = f"rucio account show {new_account}"
    exitcode, out, err = execute(command)
    assert exitcode == 0
    assert "ERROR" not in err
    assert new_account in out

    command = f"rucio account remove {new_account}"
    exitcode, out, err = execute(command)
    assert exitcode == 0
    assert "ERROR" not in err
    assert new_account not in [account["account"] for account in rucio_client.list_accounts()]

    # Test account banning
    tmp_account = account_name_generator()
    execute(f"rucio account add {tmp_account} USER")

    cmd = f"rucio account update {tmp_account} --ban"
    exitcode, _, new_ban_log = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in new_ban_log

    cmd = f"rucio account update {tmp_account} --unban"
    exitcode, _, new_unban_log = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in new_unban_log


def test_account_attribute(jdoe_account):
    fake_key = generate_uuid()[:15]
    cmd = f"rucio account attribute add {jdoe_account} --key test_{fake_key}_key --value True"
    exitcode, _, log = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in log

    cmd = f"rucio account attribute list {jdoe_account}"
    exitcode, out, err = execute(cmd)
    print(err)
    assert exitcode == 0
    assert "ERROR" not in err
    assert f"test_{fake_key}_key" in out

    cmd = f"rucio account attribute remove {jdoe_account} --key test_{fake_key}_key"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio -v account attribute list {jdoe_account}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert f"test_{fake_key}_key" not in out


def test_account_identities(rucio_client):
    tmp_account = account_name_generator()
    execute(f"rucio account add {tmp_account} USER")

    cmd = "rucio account identity list"
    exitcode, _, _ = execute(cmd)
    assert exitcode == 2  # Error raised when account is not included

    cmd = f"rucio account identity list {tmp_account}"
    _, _, err = execute(cmd)
    assert "ERROR" not in err

    email = "anon@email.com"
    cmd = f"rucio account identity add {tmp_account} --type GSS --email {email} --id {email}"
    _, _, err = execute(cmd)
    assert "ERROR" not in err
    id = [account for account in rucio_client.list_identities(tmp_account) if account["type"] == "GSS"][0]
    assert id["email"] == email
    assert id["identity"] == email

    cmd = f"rucio account identity list {tmp_account}"
    _, out, err = execute(cmd)
    assert "ERROR" not in err
    assert email in out

    cmd = f"rucio account identity add --account {tmp_account} --type NotAType --email {email} --id {email}"
    exitcode, _, _ = execute(cmd)
    assert exitcode == 2  # Fails with the argparse validation


def test_account_limit(jdoe_account, rucio_client):
    jdoe_account = jdoe_account.external
    mock_rse = "MOCK"

    bytes_limit = 10
    cmd = f"rucio account limit add {jdoe_account} --rse {mock_rse} --bytes {bytes_limit}"
    _, _, set_log = execute(cmd)
    assert "ERROR" not in set_log
    assert bytes_limit == rucio_client.get_account_limits(jdoe_account, mock_rse, locality="local")[mock_rse]

    cmd = f"rucio -v account limit list {jdoe_account} --rse {mock_rse}"
    _, out, err = execute(cmd)
    assert "ERROR" not in err
    assert mock_rse in out

    cmd = f"rucio account limit remove {jdoe_account} --rse {mock_rse}"
    _, _, rm_log = execute(cmd)
    assert "ERROR" not in rm_log
    assert rucio_client.get_account_limits(jdoe_account, mock_rse, locality="local")[mock_rse] is None

    cmd = f"rucio account limit add {jdoe_account} --rse {mock_rse} --bytes {bytes_limit} --locality NotAnOption"
    exitcode, _, _ = execute(cmd)
    assert exitcode == 2  # Fails bc locality is limited to local or global


@pytest.mark.noparallel("Changes config settings")
def test_config():
    cmd = "rucio config list"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    exitcode, _, err = execute("rucio config list --section vo-map")
    assert exitcode == 0
    assert "ERROR" not in err

    section = "vo-map"
    option = "new_option"
    value = "new_value"

    cmd = f"rucio config add --section {section} --key {option} --value {value}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    exitcode, out, err = execute("rucio config list --section vo-map")
    assert exitcode == 0
    assert "ERROR" not in err
    assert value in out

    cmd = f"rucio config remove --section {section} --key {option}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    exitcode, out, err = execute("rucio config list --section vo-map")
    assert exitcode == 0
    assert "ERROR" not in err
    assert value not in out


def test_did(rucio_client, root_account):
    scope = scope_name_generator()
    rucio_client.add_scope(account=root_account.external, scope=scope)
    dataset = file_generator().split("/")[-1]
    rucio_client.add_did(scope=scope, name=dataset, did_type="dataset")
    container = file_generator().split("/")[-1]
    rucio_client.add_did(scope=scope, name=container, did_type="container")

    cmd = f"rucio did list {scope}:*"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio did update --touch {scope}:{dataset}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio did list {scope}:* --filter type=all"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert f"{scope}:{dataset}" in out
    assert "ERROR" not in err

    cmd = f"rucio did show {scope}:{dataset}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert dataset in out  # At least list the name properly

    # Add the collection DIDs
    scope = scope_name_generator()
    rucio_client.add_scope(account=root_account.external, scope=scope)
    dataset = file_generator().split("/")[-1]

    cmd = f"rucio did add --type container {scope}:{dataset}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert dataset in [dataset for dataset in rucio_client.list_dids(scope=scope, filters=[], did_type="container")]

    scope = scope_name_generator()
    rucio_client.add_scope(account=root_account.external, scope=scope)
    dataset = file_generator().split("/")[-1]

    cmd = f"rucio did add --type dataset {scope}:{dataset}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert dataset in [dataset for dataset in rucio_client.list_dids(scope=scope, filters=[], did_type="dataset")]


def test_did_content(root_account, rucio_client):
    scope = scope_name_generator()
    rucio_client.add_scope(account=root_account.external, scope=scope)
    dataset = file_generator().split("/")[-1]
    rucio_client.add_did(scope=scope, name=dataset, did_type="dataset")
    container = file_generator().split("/")[-1]
    rucio_client.add_did(scope=scope, name=container, did_type="container")

    cmd = f"rucio did content add {scope}:{dataset} --to-did {scope}:{container}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio did content list {scope}:{container}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert dataset in out

    execute(f"rucio did content remove {scope}:{dataset} --from-did {scope}:{container}")

    cmd = f"rucio did content history {scope}:{container}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert dataset in out

    cmd = f"rucio did update --close {scope}:{container}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    # Verify it has been closed
    cmd = f"rucio did content add {scope}:{dataset} --to-did {scope}:{container}"
    _, _, err = execute(cmd)
    assert "ERROR" in err

    cmd = f"rucio did update  --open {scope}:{container}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    # Verify is has been re-opened
    cmd = f"rucio did content add {scope}:{dataset} --to-did {scope}:{container}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio did list --parent {scope}:{dataset}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert container in out

    cmd = f"rucio did content remove {scope}:{dataset} --from-did {scope}:{container}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio did list --parent {scope}:{dataset}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert container not in out  # Only checks for output not err, upstream error with mistaking claim scopes dne


def test_did_metadata(rucio_client, root_account):
    scope = scope_name_generator()
    rucio_client.add_scope(account=root_account.external, scope=scope)
    dataset = file_generator().split("/")[-1]
    rucio_client.add_did(scope=scope, name=dataset, did_type="dataset")

    metadata_value = f"mock_{generate_uuid()[:15]}"
    cmd = f"rucio did metadata add {scope}:{dataset} --key project --value {metadata_value}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert metadata_value in [value for value in rucio_client.get_metadata(scope=scope, name=dataset).values()]

    cmd = f"rucio did metadata list {scope}:{dataset}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert metadata_value in out

    # NOTE - This cannot be done in the current dataset - throws a notimplementederror
    # cmd = f"rucio -v did metadata remove --did {scope}:{dataset} --key project"
    # exitcode, _, err = execute(cmd)
    # print(err)
    # assert exitcode == 0
    # assert "ERROR" not in err
    # assert metadata_value not in [value for value in rucio_client.get_metadata(scope=scope, name=dataset).values()]


def test_upload_download():
    # DID upload/download not tested for implementation as their tests are identical to the base rucio bin versions
    cmd = "rucio upload there-is-not-a-file-here --rse mock"
    exitcode, _, _ = execute(cmd)
    assert exitcode != 2  # Failure is not due to the command structure

    cmd = "rucio download fake:fake --dir ."
    exitcode, _, _ = execute(cmd)
    assert exitcode != 2  # Failure is not due to the command structure


def test_lifetime_exception(rucio_client, mock_scope):
    from rucio.client.uploadclient import UploadClient

    input_file = tempfile.NamedTemporaryFile()
    mock_did = tempfile.NamedTemporaryFile()
    mock_rse = "MOCK-POSIX"
    upload_client = UploadClient(rucio_client)
    upload_client.upload(items=[{"path": mock_did.name, "rse": mock_rse, "did_scope": mock_scope.external}])
    with open(input_file.name, "w") as f:
        f.write(f"{mock_scope}:{mock_did.name.split('/')[-1]}")

    cmd = f"rucio lifetime-exception add -f {input_file.name} --reason mock_test -x 2100-12-30"
    exitcode, _, err = execute(cmd)
    print(err)
    assert exitcode == 0
    if "ERROR" not in err:
        assert "not affected by the lifetime model" in err
    else:
        assert "Nothing to submit" in err


@pytest.mark.dirty
def test_replica(mock_scope, rucio_client):
    mock_did = tempfile.NamedTemporaryFile()
    mock_rse = "MOCK-POSIX"

    scope = mock_scope.external
    name = mock_did.name.split("/")[-1]

    rucio_client.add_replica(mock_rse, scope, name, 1, "deadbeef")  # I don't know why this is the default adler32

    cmd = f"rucio replica list dataset {scope}:{name}"
    exitcode, out, err = execute(cmd)
    assert "ERROR" not in err
    assert exitcode == 0
    assert name not in out

    cmd = f"rucio replica list file {scope}:{name}"
    exitcode, out, err = execute(cmd)
    assert "ERROR" not in err
    assert exitcode == 0
    assert name in out

    cmd = f"rucio replica list file {scope}:{name} --pfns --rses {mock_rse}"
    exitcode, out, err = execute(cmd)
    assert "ERROR" not in err
    assert exitcode == 0
    pfn = [r for r in rucio_client.list_replicas([{"name": name, "scope": scope}])][0]["pfns"].keys()
    pfn = list(pfn)[0]
    assert pfn in out

    cmd = f"rucio replica remove {scope}:{name} --rse {mock_rse}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err


@pytest.mark.dirty
def test_replica_state(mock_scope, rucio_client):
    mock_rse = "MOCK3"
    scope = mock_scope.external

    name1 = generate_uuid()
    rucio_client.add_replica(mock_rse, mock_scope.external, name1, 4, "deadbeef")

    cmd = f"rucio replica state update bad {scope}:{name1} --rse {mock_rse} --reason testing"
    exitcode, _, err = execute(cmd)
    print(err)
    assert exitcode == 0
    if "ERROR" in err:
        assert "Details: ERROR, multiple matches" in err  # The test rses are strange. I don't know why this happens.

    name2 = generate_uuid()
    rucio_client.add_replica(mock_rse, mock_scope.external, name2, 4, "deadbeef")
    cmd = f"rucio replica state update unavailable {mock_scope}:{name2} --rse {mock_rse}  --reason testing"
    exitcode, _, err = execute(cmd)

    name3 = generate_uuid()
    rucio_client.add_replica(mock_rse, mock_scope.external, name3, 4, "deadbeef")
    cmd = f"rucio replica state update quarantine {mock_scope}:{name3} --rse {mock_rse}"
    exitcode, _, err = execute(cmd)

    assert exitcode == 0
    assert "ERROR" not in err


def test_rse(rucio_client):
    rse_name = rse_name_generator()

    cmd = f"rucio rse add {rse_name}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert rse_name in [i['rse'] for i in rucio_client.list_rses(rse_name)]

    cmd = "rucio rse list"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert rse_name in out

    cmd = f"rucio rse show {rse_name}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert rse_name in out

    value = rse_name_generator()
    cmd = f"rucio rse update {rse_name} --key city --value {value}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert value == rucio_client.list_rse_attributes(rse_name)['city']

    cmd = f"rucio rse remove {rse_name}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert rse_name not in [i for i in rucio_client.list_rses(rse_name)]


def test_rse_attribute():
    rse_name = rse_name_generator()
    _, _, err = execute(f"rucio rse add {rse_name}")
    assert "ERROR" not in err

    cmd = f"rucio rse attribute list {rse_name}"
    _, _, err = execute(cmd)
    assert "ERROR" not in err

    cmd = f"rucio rse attribute add {rse_name} --key name --value {rse_name}"
    _, _, err = execute(cmd)
    assert "ERROR" not in err


def test_rse_protocol():
    rse_name = rse_name_generator()
    _, _, err = execute(f"rucio rse add {rse_name}")
    assert "ERROR" not in err

    domain_json = """{"wan": {"read": 1, "write": 1, "delete": 1, "third_party_copy_read": 1, "third_party_copy_write": 1}}"""
    cmd = f"rucio rse protocol add {rse_name} --host-name blocklistreplica --scheme file --prefix /rucio --port 0 --impl rucio.rse.protocols.posix.Default --domain-json '{domain_json}'"
    exitcode, _, err = execute(cmd)
    print(err)
    assert "ERROR" not in err
    assert exitcode == 0

    cmd = f"rucio rse protocol remove {rse_name} --host-name blocklistreplica --scheme file"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err


def test_rse_distance():
    source_rse = "MOCK"
    dest_rse = "MOCK2"

    cmd = f"rucio rse distance remove {source_rse} {dest_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    if "ERROR" in err:
        assert f"Distance from {source_rse} to {dest_rse}" in err

    cmd = f"rucio rse distance add {source_rse} {dest_rse} --distance 1"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio rse distance show {source_rse} {dest_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert dest_rse in out
    assert "1" in out

    cmd = f"rucio rse distance update {source_rse} {dest_rse} --distance 10"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio rse distance show {source_rse} {dest_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert dest_rse in out
    assert "10" in out

    cmd = f"rucio rse distance remove {source_rse} {dest_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err


def test_rse_limits(rucio_client):
    mock_rse = 'MOCK'
    limit = "mock_limit"

    cmd = f"rucio rse limit add {mock_rse} --limit {limit} 100"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert 100 == rucio_client.get_rse_limits(mock_rse)[limit]

    cmd = f"rucio rse limit remove {mock_rse} --limit {limit}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    try:
        assert limit not in rucio_client.get_rse_limits(mock_rse).keys()
    except RucioException:  # Can throw an error if the mock_rse has no limits
        pass


def test_rse_qos_policy(rucio_client):
    mock_rse = "MOCK"
    policy = "SOMETHING_I_GUESS"

    cmd = f"rucio rse qos add {mock_rse} --policy {policy}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert policy in rucio_client.list_qos_policies(mock_rse)

    cmd = f"rucio rse qos list {mock_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert policy in out

    cmd = f"rucio rse qos remove {mock_rse} --policy {policy}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert policy not in rucio_client.list_qos_policies(mock_rse)


@pytest.mark.dirty
def test_rule(rucio_client, mock_scope):
    mock_rse = "MOCK-POSIX"
    rule_rse = "MOCK"

    scope = mock_scope.external
    name = generate_uuid()
    rucio_client.add_replica(rse=mock_rse, scope=scope, name=name, bytes_=4, adler32="deadbeef")

    cmd = f"rucio rule add {scope}:{name} --copies 1 --rses {rule_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    rule_id = out.strip("\n")

    cmd = f"rucio rule list --did {scope}:{name}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert rule_id in out

    # Below functionality is questionable - listing by ID works _sometimes_ if the database has been updated with the new rule

    # cmd = f"rucio -v rule list --rule-id {rule_id}"
    # exitcode, out, err = execute(cmd)
    # assert exitcode == 0
    # assert "ERROR" not in err
    # assert rule_id in out

    move_rse = "MOCK2"
    cmd = f"rucio rule move {rule_id} --rses {move_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio rule update {rule_id} --priority 3"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    # Check that you can lock and unlock a rule
    cmd = f"rucio rule update {rule_id} --locked False"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert not rucio_client.get_replication_rule(rule_id)['locked']

    cmd = f"rucio -v rule update {rule_id} --locked True"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert rucio_client.get_replication_rule(rule_id)['locked']

    # Testing the two different lifetime type options
    cmd = f"rucio rule update {rule_id} --lifetime 10"
    exitcode, out, err = execute(cmd)
    print(out, err)
    assert exitcode == 0
    assert "ERROR" not in err
    rule_info = rucio_client.get_replication_rule(rule_id)
    assert abs(rule_info["updated_at"] - rule_info["expires_at"]).seconds == 10

    cmd = f"rucio rule update {rule_id} --lifetime None"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert rucio_client.get_replication_rule(rule_id)["expires_at"] is None

    # Do one without a child rule so i can delete it
    new_name = generate_uuid()
    rucio_client.add_replica(mock_rse, scope, new_name, 4, "deadbeef")
    cmd = f"rucio rule add {scope}:{new_name} --copies 1 --rses {mock_rse}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    rule_id = out.strip("\n")

    cmd = f"rucio rule remove {rule_id}"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err


def test_scope():
    new_scope = scope_name_generator()
    cmd = f"rucio scope add {new_scope} --account root"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = "rucio scope list --account root"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    # assert new_scope in out
    # See issue https://github.com/rucio/rucio/issues/7316


def test_subscription(rucio_client, mock_scope):
    subscription_name = generate_uuid()

    filter_ = json.dumps({})
    rules = json.dumps([{"copies": 1, "rse_expression": "JDOE_DATADISK", "lifetime": 3600, "activity": "User Subscriptions"}])

    cmd = f"rucio subscription add {subscription_name} --account root --filter '{filter_}' --rule '{rules}'"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    cmd = f"rucio subscription show {subscription_name} --account root"
    exitcode, out, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err
    assert subscription_name in out

    did = [i for i in rucio_client.list_dids(mock_scope.external, filters=[{}], did_type="all")][0]
    cmd = f"rucio subscription touch {mock_scope.external}:{did}"
    exitcode, _, err = execute(cmd)
    assert exitcode == 0
    assert "ERROR" not in err

    filter_ = json.dumps({"did_type": "all"})
    rule = json.dumps({})
    cmd = f"rucio subscription update {subscription_name} --filter '{filter_}' --rule {rule}"
    exitcode, _, err = execute(cmd)
    print(err)
    assert exitcode == 0
    assert "ERROR" not in err
